# Generated by Django 4.1.2 on 2022-12-09 00:04

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('sphinxhostingcore', '0003_globaltoc'),
    ]

    operations = [
        migrations.AlterField(
            model_name='sphinxpage',
            name='next_page',
            field=models.ForeignKey(help_text='The next page in the documentation set', null=True, on_delete=django.db.models.deletion.CASCADE, related_name='previous_page', to='sphinxhostingcore.sphinxpage'),
        ),
    ]



from argparse import ArgumentParser

from django.core.management.base import BaseCommand

from sphinx_hosting.exc import VersionAlreadyExists
from sphinx_hosting.models import Project
from sphinx_hosting.importers import SphinxPackageImporter


class Command(BaseCommand):
    """
    Import a Sphinx documentation tarfile into the database.

    We will use the the slugified version (converted to only letters, numbers,
    dashes and underscores) of the ``project`` variable from the Sphinx
    ``conf.py`` that built the tarfile as the ``machine_name`` of the
    :py:class:`Project` to load into, and we will use the "release" variable
    from ``conf.py`` to build the :py:class:`Version` object to load into.

    The Sphinx documentation tarfile should have been built via the``json``
    output from ``sphinx-build``, either::

        make json

    or::

        sphinx-build -n -b json build/json

    Then the tarfile should be built like so::

        cd build
        tar zcf mydocs.tar.gz json

    The files in the resulting tarfile should be contained in a folder.  We
    something like::

        json/py-modindex.fjson
        json/globalcontext.json
        json/_static
        json/last_build
        json/genindex.fjson
        json/objectstore.fjson
        json/index.fjson
        json/environment.pickle
        json/searchindex.json
        json/objects.inv
        ...

    """
    args = '<tarfile>'
    help = 'Imports a tarfile of Sphinx documentation into a pre-existing project.'

    def add_arguments(self, parser: ArgumentParser) -> None:
        parser.add_argument(
            'tarfile',
            metavar='tarfile',
            type=str,
            help='The filename of the tarfile to import.'
        )
        parser.add_argument(
            '--force',
            action='store_true',
            default=False,
            help='Overwrite documentation for any matching existing version.'
        )

    def handle(self, *args, **options) -> None:
        importer = SphinxPackageImporter()
        try:
            version = importer.run(filename=options['tarfile'], force=options['force'])
        except VersionAlreadyExists as e:
            self.stdout.write(self.style.ERROR(str(e)))
        except Project.DoesNotExist as e:
            self.stdout.write(self.style.ERROR(str(e)))
        self.stdout.write(
            self.style.SUCCESS(
                f'Imported to Version(id={version.id}) of '
                f'Project(id={version.project.id}, title={version.project.title})'
            )
        )



from argparse import ArgumentParser
from typing import cast

from django.core.management.base import BaseCommand
from django.utils.text import slugify
from rich.tree import Tree
from rich import print as rich_print

from sphinx_hosting.models import SphinxPage, Version, TreeNode


class TreePrinter:
    """
    Parse the tree of :py:class:`sphinx_hosting.models.TreeNode` objects
    we get from :py:meth:`sphinx_hosting.models.Version.page_tree` and
    print them to stdout in a pretty format.
    """

    def __init__(self, version: Version):
        self.sphinx_tree = version.page_tree
        self.tree: Tree = Tree(self.title(self.sphinx_tree.head))
        self.build(self.tree, cast(TreeNode, self.sphinx_tree.head))

    def title(self, node: TreeNode) -> str:
        return f'{node.title} [{cast(SphinxPage, node.page).id}]'

    def build(self, branch: Tree, node: TreeNode):
        if node.children:
            for child in node.children:
                leaf = branch.add(self.title(child))
                self.build(leaf, child)

    def print(self):
        rich_print(self.tree)


class Command(BaseCommand):
    """
    **Usage**: ``./manage.py print_doctree <project_machine_name> <version number>``

    Print the page tree for a :py:class:`Version`
    """
    args = '<tarfile>'
    help = 'Print the deduced page hierarchy for a version of a project.'

    def add_arguments(self, parser: ArgumentParser) -> None:
        parser.add_argument(
            'project',
            metavar='project_machine_name',
            type=str,
            help='The machine_name of the project.'
        )
        parser.add_argument(
            'version',
            metavar='version_number',
            type=str,
            help='The version number for which we want to print a page tree.'
        )

    def handle(self, *args, **options) -> None:
        try:
            version = Version.objects.get(
                project__machine_name=slugify(options['project']),
                version=options['version']
            )
        except Version.DoesNotExist as e:
            self.stdout.write(self.style.ERROR(str(e)))

        self.stdout.write(self.style.SUCCESS(f'\n{version.project.title}: {version.version}\n'))
        TreePrinter(version).print()



from typing import List, Type

from django.db.models import Model, QuerySet
from haystack import indexes

from .models import SphinxPage


class SphinxPageIndex(indexes.SearchIndex, indexes.Indexable):

    text = indexes.CharField(document=True, use_template=True)
    project_id = indexes.CharField(model_attr='version__project__id', faceted=True)
    project = indexes.CharField(model_attr='version__project__machine_name')
    version_id = indexes.CharField(model_attr='version__id')
    classifiers = indexes.MultiValueField(faceted=True)
    modified = indexes.DateTimeField(model_attr='modified')
    is_latest = indexes.CharField(faceted=True)

    def get_model(self) -> Type[Model]:
        return SphinxPage

    def prepare_is_latest(self, obj: SphinxPage) -> str:
        """
        Elasticsearch doesn't know how to deal with Python boolean values, so we set
        a string key on our document to either ``true`` or ``false`` depending on
        whether this page is part of the most recent version for a project or not.

        Args:
            obj: The page we're indexing

        Returns:
            Either ``true`` or ``false``.
        """
        is_latest = obj.version == obj.version.project.latest_version
        if is_latest:
            return 'true'
        return 'false'

    def prepare_classifiers(self, obj: SphinxPage) -> List[str]:
        return [classifier.name for classifier in obj.version.project.classifiers.all()]

    def index_queryset(self, using=None) -> QuerySet:
        """
        Used when the entire index for model is updated.
        """
        return self.get_model().objects.filter(searchable=True)



from django import forms

from .validators import validate_machine_name


class MachineNameField(forms.SlugField):
    """
    This is a form field for our
    :py:class:`sphinx_hosting.fields.MachineNameField` that applies the
    appropriate validators.

    The difference this field and :py:class:`django.forms.SlugField` is that
    this field will allow "-" characters in the value
    """
    default_validators = [validate_machine_name]



import re
import urllib.parse

from django.core.management.base import BaseCommand
import lxml.html

from sphinx_hosting.models import SphinxPage


class Command(BaseCommand):
    """
    **Usage**: ``./manage.py fix_broken_hrefs``

    This is a one-shot command to fix broken hrefs from imported data that
    used ``django-sphinx-hosting`` version 1.1.5 or earlier.  This command
    fixes internal hrefs that were imported to be template tags which resolve
    the proper page at render time.

    Print the page tree for a :py:class:`Version`
    """

    def _fix_link_hrefs(self, body: str, project_machine_name: str, version: str) -> str:
        """
        Given an HTML body of a Sphinx page, update the ``<a href="path">``
        references to be absolute.  If we don't do this, any links on the
        index page the version will be relative to the index page, instead of
        being relative to the root of the docs, and won't work.

        Args:
            body: the HTML body of a Sphinx document

        Returns:
            ``body`` with its ``<a>`` urls and updated
        """
        html = lxml.html.fromstring(body)
        links = html.cssselect('a.reference.internal')
        for link in links:
            href = link.attrib['href']
            anchor = None
            if '#' in href:
                href, anchor = href.split('#')
            if href.endswith('/'):
                href = href[:-1]
            href = re.sub('^(../)*', '', href)
            link.attrib['href'] = "{{% url 'sphinx_hosting:sphinxpage--detail' project_slug='{}' version='{}' path='{}' %}}".format(
                project_machine_name,
                version,
                href
            )
            if anchor:
                link.attrib['href'] += f'#{anchor}'
        body = lxml.html.tostring(html).decode('utf-8')
        tags = [m.group() for m in re.finditer(r'%7B%%20.*?%20%%7D', body)]
        for tag in tags:
            body = body.replace(tag, urllib.parse.unquote(tag))
        return body

    def handle(self, *args, **options) -> None:
        for page in SphinxPage.objects.all():
            if page.relative_path not in SphinxPage.SPECIAL_PAGES:
                if not page.body:
                    continue
                page.body = self._fix_link_hrefs(
                    page.body,
                    page.version.project.machine_name,
                    page.version.version
                )
                page.save()
                print(
                    f'Fixed page {page.version.project.machine_name}-{page.version.version}:'
                    f'{page.title}'
                )



from typing import List, Optional, cast

from crequest.middleware import CrequestMiddleware
from django.contrib.auth.models import AbstractUser
from django.http.request import HttpRequest
from django.urls import reverse_lazy
from wildewidgets import (
    Block,
    BreadcrumbBlock,
    BreadcrumbItem,
    LinkedImage,
    Menu,
    MenuItem,
    TablerVerticalNavbar
)

from ..settings import (
    LOGO_IMAGE,
    LOGO_URL,
    LOGO_WIDTH,
    SITE_NAME
)
from .search import GlobalSearchFormWidget


class SphinxHostingMainMenu(Menu):
    """
    This is the primary menu that appears in :py:class:`SphinxHostingSidebar`
    directly underneath the "Search" form,
    :py:class:`sphinx_hosting.wildewidgets.search.GlobalSearchFormWidget`.

    It gives access to all the views that normal, non privileged users should be
    allowed to use.
    """
    title: str = "Main"
    items = [
        MenuItem(
            text='Projects',
            icon='bookshelf',
            url=reverse_lazy('sphinx_hosting:project--list')
        )
    ]

    def __init__(self, *items, **kwargs) -> None:
        self.active_item: Optional[str] = None
        super().__init__(*items, **kwargs)

    def build_menu(self, items: List[MenuItem]) -> None:
        """
        Programatically build our menu here.  We have this code because the
        presence of some items in this menu depends on the user's permissions.

        We have to do it here because if we do it in ``__init__``, that's too
        early in the Django boostrap and the global Django ``urlpatterns`` have
        not finished building, and even ``reverse_lazy`` fails.
        """
        request: HttpRequest = CrequestMiddleware.get_request()
        user: AbstractUser = cast(AbstractUser, request.user)
        if user.has_perm('sphinxhostingcore.view_classifier'):
            item = MenuItem(
                text='Classifiers',
                icon='bookshelf',
                url=reverse_lazy('sphinx_hosting:classifier--index')
            )
            items.append(item)
        for item in items:
            if self.active_item is not None:
                item.set_active(self.active_item)
        super().build_menu(items)

    def activate(self, text: str) -> bool:
        """
        Normally, how activate works is that it looks through our menu items,
        finds the one whose ``text`` or ``url`` matches ``text``.

        In our case, we're building the menu items in :py:meth:`build_menu`,
        which occurs after :py:meth:`activate` would be called, so we have to
        save the ``text`` here, and use it later in :py:meth:`build_menu`.

        Args:
            text: the text to search for among our :py:attr:`items`

        Returns:
            We always return ``True`` here, since we won't know if we match
            until later.
        """
        self.active_item = text
        return True


class SphinxHostingSidebar(TablerVerticalNavbar):
    """
    This is the vertical menu area on the left of the page.  It houses our
    search form,
    :py:class:`sphinx_hosting.wildewidgets.search.GlobalSearchFormWidget`, our
    main menu :py:class:`SphinxHostingMainMenu`, possibly our utility menu
    :py:class:`SphinxHostingLookupsMenu` and finally the global navigation
    for a :py:class:`sphinx_hosting.models.Version` when we're reading our
    documentation.
    """

    hide_below_viewport: str = 'xl'
    branding = Block(
        LinkedImage(
            image_src=LOGO_IMAGE,
            image_width=LOGO_WIDTH,
            image_alt=SITE_NAME,
            css_class='d-flex justify-content-center',
            url=LOGO_URL
        ),
        GlobalSearchFormWidget(css_class='ms-auto ms-xl-0 align-self-center mt-3'),
        css_class='d-flex flex-row flex-xl-column justify-content-between flex-grow-1 flex-xl-grow-0'
    )
    contents = [
        SphinxHostingMainMenu(),
    ]
    wide: bool = True


class SphinxHostingBreadcrumbs(BreadcrumbBlock):

    items: List[BreadcrumbItem] = [
        BreadcrumbItem(
            title=SITE_NAME,
            url=reverse_lazy('sphinx_hosting:project--list')
        )
    ]



from django.db import models
from haystack import signals

from .models import SphinxPage


class SphinxHostingSignalProcessor(signals.BaseSignalProcessor):

    def setup(self):
        # Listen only to the ``User`` model.
        models.signals.post_save.connect(self.handle_save, sender=SphinxPage)
        models.signals.post_delete.connect(self.handle_delete, sender=SphinxPage)

    def teardown(self):
        # Disconnect only for the ``User`` model.
        models.signals.post_save.disconnect(self.handle_save, sender=SphinxPage)
        models.signals.post_delete.disconnect(self.handle_delete, sender=SphinxPage)



from typing import Type, cast

from crispy_forms.helper import FormHelper
from crispy_forms.layout import Submit, Layout, Field, Fieldset, HTML, ButtonHolder

from crispy_forms.bootstrap import FieldWithButtons
from django import forms
from django.db.models import Model
from django.urls import reverse, reverse_lazy
from haystack.forms import SearchForm

from .models import Project, ProjectRelatedLink


class GlobalSearchForm(SearchForm):
    """
    This is the search form at the top of the sidebar, underneath the logo.  It is
    a subclass of :py:class:`haystack.forms.SearchForm`, and does a search of our
    haystack backend.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.helper = FormHelper()
        self.helper.form_class = 'form-horizontal px-3'
        self.helper.form_method = 'get'
        self.helper.form_show_labels = False
        # This is a reverse_lazy on purpose because the form gets instantiated as
        # the urlpatterns are being made
        self.helper.form_action = reverse_lazy('sphinx_hosting:search')
        self.helper.layout = Layout(
            FieldWithButtons(
                Field('q', css_class='text-dark', placeholder='Search'),
                HTML('<button type="submit" class="btn btn-primary"><span class="bi bi-search"></span></button>')
            )
        )


class ProjectCreateForm(forms.ModelForm):
    """
    This is the form we use to create a new
    :py:class:`sphinx_hosting.models.Project`.  The difference between this and
    :py:class:`sphinx_hosting.forms.ProjectUpdateForm` is that the user can set
    :py:attr:`sphinx_hosting.models.Project.machine_name` here, but can't in
    :py:class:`sphinx_hosting.forms.ProjectUpdateForm`.  ``machine_name`` should not change after the
    project is created.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.helper = FormHelper()
        self.helper.form_class = 'form-horizontal'
        self.helper.label_class = 'col-lg-3'
        self.helper.field_class = 'col'
        self.helper.form_method = 'post'
        self.helper.form_action = reverse('sphinx_hosting:project--create')
        self.helper.layout = Layout(
            Fieldset(
                '',
                Field('title'),
                Field('machine_name'),
                Field('description'),
            ),
            ButtonHolder(
                Submit('submit', 'Save', css_class='btn btn-primary'),
                css_class='d-flex flex-row justify-content-end button-holder'
            )
        )

    class Meta:
        model: Type[Model] = Project
        exclude = (
            # These are relational fields that will be handled separately
            'versions',
            'classifiers',
            'permission_groups',

            # These are maintained automatically
            'created',
            'modified'
        )
        widgets = {
            'description': forms.Textarea(attrs={'cols': 50, 'rows': 3}),
        }


class ProjectUpdateForm(forms.ModelForm):
    """
    This is the form we use to update an existing
    :py:class:`sphinx_hosting.models.Project`.  The difference between this and
    :py:class:`.ProjectCreateForm` is that the user cannot change
    :py:attr:`sphinx_hosting.models.Project.machine_name` here, but can in
    :py:class:`ProjectCreateForm`.  ``machine_name`` should not change after the
    project is created.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.helper = FormHelper()
        self.helper.form_class = 'form-horizontal'
        self.helper.label_class = 'col-lg-3'
        self.helper.field_class = 'col'
        self.helper.form_method = 'post'
        self.helper.form_action = reverse('sphinx_hosting:project--update', args=[self.instance.machine_name])
        self.helper.layout = Layout(
            Fieldset(
                '',
                Field('title'),
                Field('description'),
            ),
            ButtonHolder(
                Submit('submit', 'Save', css_class='btn btn-primary'),
                css_class='d-flex flex-row justify-content-end button-holder'
            )
        )

    class Meta:
        model: Type[Model] = Project
        fields = (
            'title',
            'description',
        )
        widgets = {
            'description': forms.Textarea(attrs={'cols': 50, 'rows': 3}),
        }


class ProjectReadonlyUpdateForm(forms.ModelForm):
    """
    This is the form we use to on the
    :py:class:`sphinx_hosting.views.ProjectDetailView` to show the viewer the
    project title and description.  The difference between this and
    :py:class:`ProjectUpdateForm` is that all the fields are readonly, and there
    are no submit buttons.  We're doing it this way instead of just rendering a
    non-form widget so that we can ensure that the page looks the same.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.fields['title'].widget.attrs['readonly'] = True
        self.helper = FormHelper()
        self.helper.form_class = 'form-horizontal'
        self.helper.label_class = 'col-lg-3'
        self.helper.field_class = 'col'
        self.helper.form_method = 'get'
        self.helper.layout = Layout(
            Fieldset(
                '',
                Field('title'),
                Field('description'),
            )
        )

    class Meta:
        model: Type[Model] = Project
        fields = (
            'title',
            'description',
        )
        widgets = {
            'description': forms.Textarea(
                attrs={
                    'cols': 50,
                    'rows': 3,
                    'readonly': 'readonly',
                }
            ),
        }


class ProjectRelatedLinkBaseForm(forms.ModelForm):

    def __init__(self, *args, project_machine_name: str = None, **kwargs):
        super().__init__(*args, **kwargs)
        self.helper = FormHelper()
        self.helper.form_class = 'form-horizontal'
        self.helper.label_class = 'col-lg-3'
        self.helper.field_class = 'col'
        self.helper.form_method = 'post'
        self.helper.layout = Layout(
            Fieldset(
                '',
                Field('title'),
                Field('uri'),
            ),
            ButtonHolder(
                Submit('submit', 'Save', css_class='btn btn-primary'),
                css_class='d-flex flex-row justify-content-end button-holder'
            )
        )

    class Meta:
        model: Type[Model] = ProjectRelatedLink
        fields = (
            'title',
            'uri',
        )


class ProjectRelatedLinkCreateForm(ProjectRelatedLinkBaseForm):

    def __init__(self, *args, project: Project = None, **kwargs):
        super().__init__(*args, **kwargs)
        self.helper.form_action = reverse(
            'sphinx_hosting:projectrelatedlink--create',
            args=[cast(Project, project).machine_name]
        )


class ProjectRelatedLinkUpdateForm(ProjectRelatedLinkBaseForm):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.helper.form_action = self.instance.get_update_url()


class VersionUploadForm(forms.Form):
    """
    This is the form on :py:class:`sphinx_hosting.views.ProjectDetailView` that
    allows the user to upload a new documentation set.

    Keyword Args:
        project: the project to which this documentation set should be
            associated
    """

    file = forms.FileField(label='')

    def __init__(self, *args, project: Project = None, **kwargs):
        super().__init__(*args, **kwargs)
        self.helper = FormHelper()
        self.helper.form_class = 'form'
        self.helper.form_method = 'post'
        if project:
            self.helper.form_action = reverse(
                'sphinx_hosting:version--upload', kwargs={'slug': project.machine_name}
            )
        self.helper.layout = Layout(
            Fieldset(
                '',
                Field('file'),
            ),
            ButtonHolder(
                Submit('submit', 'Import', css_class='btn btn-primary'),
                css_class='d-flex flex-row justify-content-end button-holder'
            )
        )



# Generated by Django 4.1.2 on 2023-01-19 19:21

from django.db import migrations
from django.core.management import call_command

fixture: str = 'classifiers'


def load_fixture(apps, schema_editor):
    call_command('loaddata', fixture, app_label='sphinxhostingcore')


def unload_fixture(apps, schema_editor):
    Classifier = apps.get_model("sphinxhostingcore", "Classifier")
    Classifier.objects.all().delete()


class Migration(migrations.Migration):

    dependencies = [
        ('sphinxhostingcore', '0006_classifiers_and_permissiongroups'),
    ]

    operations = [
        migrations.RunPython(load_fixture, reverse_code=unload_fixture),
    ]



from rest_framework import permissions


class ChangeProjectPermission(permissions.BasePermission):
    """
    Check to see if the user can change projects.

    .. note::
        This is only for use in on viewsets that are not actually the
        :py:class:`sphinx_hosting.api.models.ProjectViewSet`.  For that one, just use
        :py:class:`restramework.permissions.DjangoModelPermissions`.
    """

    def has_permission(self, request, view):
        if request.method in permissions.SAFE_METHODS:
            return True
        if request.user.has_perm('sphinxhostingcore.change_project'):
            return True
        return False


class AddVersionPermission(permissions.BasePermission):
    """
    Check to see if the user can add versions.

    .. note::
        This is really only useful on
        :py:class:`sphinx_hosting.api.models.VersionUploadView`.
    """

    def has_permission(self, request, view):
        if request.method in permissions.SAFE_METHODS:
            return True
        if request.user.has_perm('sphinxhostingcore.add_version'):
            return True
        return False



from django import template

from ..models import SphinxImage

register = template.Library()


@register.simple_tag
def sphinximage_url(pk: int):
    """
    Return the URL to the :py:class:`sphinx_hosting.models.SphinxImage` identified
    by the primary key ``pk``

    Args:
        pk: the primary key of the ``SphinxImage`` object

    Returns:
        The URL for the image file
    """
    return SphinxImage.objects.get(pk=pk).file.url



# Generated by Django 4.1.2 on 2023-02-03 23:27

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('sphinxhostingcore', '0007_load_classifiers'),
    ]

    operations = [
        migrations.AddField(
            model_name='version',
            name='archived',
            field=models.BooleanField(default=False, help_text='Whether this version should be excluded from search indexes', verbose_name='Archived?'),
        ),
    ]



import os
import tempfile
from typing import Dict, List, Optional, Type, cast

from braces.views import (
    FormInvalidMessageMixin,
    FormValidMessageMixin,
    LoginRequiredMixin,
    MessageMixin,
    PermissionRequiredMixin,
)
from django.contrib import messages
from django.contrib.auth.models import AbstractUser
from django.core.files.storage import FileSystemStorage
from django.db.models import Model, QuerySet
from django.forms import ModelForm, Form
from django.http import HttpResponse, HttpRequest, Http404
from django.shortcuts import redirect, get_object_or_404
from django.utils.translation import gettext as _
from django.urls import reverse, reverse_lazy
from django.views.generic.edit import (
    BaseFormView,
    BaseCreateView,
    BaseUpdateView
)
from django.views.generic import (
    CreateView,
    DeleteView,
    DetailView,
    TemplateView,
    UpdateView
)
from haystack.forms import ModelSearchForm
from haystack.query import SearchQuerySet
from haystack.generic_views import SearchView
from wildewidgets import (
    Navbar,
    NavbarMixin,
    StandardWidgetMixin,
    WidgetListLayout,
    Widget,
    WidgetStream
)
from wildewidgets.viewsets import ModelViewSet

from .forms import (
    ProjectCreateForm,
    ProjectUpdateForm,
    ProjectReadonlyUpdateForm,
    ProjectRelatedLinkCreateForm,
    ProjectRelatedLinkUpdateForm,
    VersionUploadForm
)
from .importers import SphinxPackageImporter
from .logging import logger
from .models import (
    Classifier,
    Project,
    ProjectRelatedLink,
    Version,
    SphinxPage
)

from .wildewidgets import (
    ProjectClassifierListWidget,
    ProjectClassifierSelectorWidget,
    ProjectCreateModalWidget,
    ProjectDetailWidget,
    ProjectInfoWidget,
    ProjectRelatedLinkCreateModalWidget,
    ProjectRelatedLinksWidget,
    ProjectRelatedLinksListWidget,
    ProjectVersionsTableWidget,
    ProjectTableWidget,
    PagedSearchLayout,
    SphinxPageGlobalTableOfContentsMenu,
    SphinxHostingBreadcrumbs,
    SphinxHostingSidebar,
    SphinxPageLayout,
    VersionInfoWidget,
    VersionSphinxPageTableWidget,
    VersionSphinxImageTableWidget,
    VersionUploadBlock,
)

# ===========================
# Viewsets
# ===========================


class ClassifierViewSet(ModelViewSet):
    model = Classifier
    template_name = 'sphinx_hosting/base.html'
    breadcrumbs_class = SphinxHostingBreadcrumbs
    navbar_class = SphinxHostingSidebar


# ===========================
# View Mixins
# ===========================

class SphinxHostingMenuMixin(NavbarMixin):

    navbar_class: Type[Navbar] = SphinxHostingSidebar


class WildewidgetsMixin(StandardWidgetMixin):  # pylint: disable=abstract-method
    """
    We subclass :py:class:`wildewidgets.StandardWidgetMixin` here so that we can define our
    standard template.
    """
    template_name = 'sphinx_hosting/base.html'


# ===========================
# Projects
# ===========================


class ProjectListView(
    LoginRequiredMixin,
    WildewidgetsMixin,
    SphinxHostingMenuMixin,
    TemplateView
):
    """
    This is the view we use for the home page of the ``sphinx_hosting`` app.  It
    lists all available :py:class:`sphinx_hosting.models.Project` objects from
    the database and allows the user to create new Projects.
    """
    menu_item: str = 'Projects'

    def get_content(self) -> Widget:
        layout = WidgetStream()
        user = cast(AbstractUser, self.request.user)
        layout.add_widget(ProjectTableWidget(user))
        if user.has_perm('sphinxhostingcore.add_project'):
            layout.add_widget(ProjectCreateModalWidget())
        return layout

    def get_breadcrumbs(self) -> SphinxHostingBreadcrumbs:
        breadcrumbs = SphinxHostingBreadcrumbs()
        return breadcrumbs


class ProjectDetailView(
    LoginRequiredMixin,
    WildewidgetsMixin,
    SphinxHostingMenuMixin,
    DetailView
):
    """
    This view handles displaying the details page for a
    :py:class:`sphinx_hosting.models.Project` and handles updates to the project
    settings itself.
    """

    menu_item: str = "Projects"

    model: Type[Model] = Project
    slug_field: str = 'machine_name'

    def get_content(self) -> Widget:
        layout = WidgetListLayout(self.object.title)
        layout.add_widget(ProjectInfoWidget(self.object))
        layout.add_widget(
            ProjectDetailWidget(
                self.object,
                form=ProjectReadonlyUpdateForm(instance=self.object)
            )
        )
        layout.add_widget(ProjectRelatedLinksListWidget(queryset=self.object.related_links.all()))
        layout.add_widget(ProjectClassifierListWidget(queryset=self.object.classifiers.all()))
        layout.add_widget(ProjectVersionsTableWidget(project_id=self.object.pk))
        user = cast(AbstractUser, self.request.user)
        version = self.object.latest_version
        if version and version.head:
            layout.add_sidebar_link_button(
                'Read Docs',
                reverse(
                    'sphinx_hosting:sphinxpage--detail',
                    args=[
                        self.object.machine_name,
                        version.version,
                        version.head.relative_path
                    ]
                ),
                color='orange',
                css_class='mb-3'
            )
        if user.has_perm('sphinxhostingcore.change_project'):
            layout.add_sidebar_link_button(
                'Edit Project',
                reverse('sphinx_hosting:project--update', args=[self.object.machine_name]),
                color='azure',
            )
        return layout

    def get_breadcrumbs(self) -> SphinxHostingBreadcrumbs:
        breadcrumbs = SphinxHostingBreadcrumbs()
        breadcrumbs.add_breadcrumb(self.object.title)
        return breadcrumbs


class ProjectCreateView(
    LoginRequiredMixin,
    PermissionRequiredMixin,
    FormInvalidMessageMixin,
    FormValidMessageMixin,
    CreateView
):
    """
    This is a POST only view for creating a
    :py:class:`sphinx_hosting.models.Project`.  The POST will be submitted from
    the "Add Project" modal form displayed on :py:class:`ProjectListView`.
    """
    permission_required: str = 'sphinxhostingcore.add_project'
    model: Type[Model] = Project
    form_class: Type[ModelForm] = ProjectCreateForm

    def get_form_valid_message(self) -> str:
        obj = cast(Project, self.object)
        return f'Added Project "{obj.title}"'

    def form_invalid(self, form: ModelForm) -> HttpResponse:
        """
        If the form is invalid, we want to display the errors to the user and
        redirect them back to the project list page.

        Args:
            form: The form that was submitted

        Returns:
            The redirect response
        """
        for k, errors in form.errors.as_data().items():
            for error in errors:
                messages.error(self.request, f"{k}: {error.message}")
        return redirect('sphinx_hosting:project--list')

    def get_success_url(self) -> str:
        obj = cast(Project, self.object)
        logger.info("project.create.success project=%s id=%s", obj.machine_name, obj.id)
        return reverse('sphinx_hosting:project--update', args=[obj.machine_name])

    def get_form_invalid_message(self) -> str:
        logger.warning('project.create.failed.validation')
        return _("Couldn't create this project due to validation errors; see below.")


class ProjectUpdateView(
    LoginRequiredMixin,
    PermissionRequiredMixin,
    FormValidMessageMixin,
    WildewidgetsMixin,
    SphinxHostingMenuMixin,
    UpdateView
):
    """
    This view handles displaying the details page for a
    :py:class:`sphinx_hosting.models.Project` and handles updates to the project
    settings itself.
    """

    menu_item: str = "Projects"

    permission_required: str = 'sphinxhostingcore.change_project'
    model: Type[Model] = Project
    form_class: Type[ModelForm] = ProjectUpdateForm
    slug_field: str = 'machine_name'

    def get_content(self) -> Widget:
        layout = WidgetListLayout(self.object.title)
        layout.add_widget(ProjectInfoWidget(self.object))
        layout.add_widget(ProjectDetailWidget(self.object))
        layout.add_widget(ProjectRelatedLinksWidget(self.object))
        layout.add_widget(ProjectClassifierSelectorWidget(self.object))
        layout.add_widget(ProjectVersionsTableWidget(project_id=self.object.pk))
        layout.add_widget(ProjectRelatedLinkCreateModalWidget(self.object))
        version = self.object.latest_version
        user = cast(AbstractUser, self.request.user)
        if version and version.head:
            layout.add_sidebar_link_button(
                'Read Docs',
                reverse(
                    'sphinx_hosting:sphinxpage--detail',
                    args=[
                        self.object.machine_name,
                        version.version,
                        version.head.relative_path
                    ]
                ),
                color='orange',
                css_class='mb-3'
            )
        if user.has_perm('sphinxhostingcore.add_version'):
            layout.add_sidebar_bare_widget(
                VersionUploadBlock(form=VersionUploadForm(project=self.object))
            )
        if user.has_perm('sphinxhostingcore.delete_project'):
            layout.add_sidebar_form_button(
                'Delete Project',
                reverse('sphinx_hosting:project--delete', args=[self.object.machine_name]),
                color='outline-secondary',
                confirm_text=_("Are you sure you want to delete this project?"),
            )
        return layout

    def get_breadcrumbs(self) -> SphinxHostingBreadcrumbs:
        breadcrumbs = SphinxHostingBreadcrumbs()
        breadcrumbs.add_breadcrumb(self.object.title)
        return breadcrumbs

    def get_form_valid_message(self) -> str:
        return f'Updated Project "{self.object.title}"'

    def get_success_url(self) -> str:
        logger.info('project.update.success project=%s id=%s', self.object.machine_name, self.object.id)
        return reverse('sphinx_hosting:project--update', args=[self.object.machine_name])


class ProjectDeleteView(
    LoginRequiredMixin,
    PermissionRequiredMixin,
    DeleteView
):
    model: Type[Model] = Project
    permission_required: str = 'sphinxhostingcore.delete_project'
    success_url = reverse_lazy('sphinx_hosting:project--list')
    slug_field: str = 'machine_name'

    def delete(self, request: HttpRequest, *args, **kwargs) -> HttpResponse:
        response = super().delete(request, *args, **kwargs)
        logger.info(
            "project.delete.success project=%s id=%s",
            self.object.machine_name,
            self.object.pk
        )
        return response


# ===========================
# ProjectRelatedLinks
# ===========================

class ProjectRelatedLinkCreateView(
    LoginRequiredMixin,
    PermissionRequiredMixin,
    FormValidMessageMixin,
    BaseCreateView
):
    form_class = ProjectRelatedLinkCreateForm
    model: Type[Model] = ProjectRelatedLink
    permission_required: str = 'sphinxhostingcore.change_project'

    def get_form_kwargs(self) -> Dict[str, str]:
        self.project: Project = get_object_or_404(Project, machine_name=self.kwargs['project_slug'])
        kwargs = super().get_form_kwargs()
        kwargs['project'] = self.project
        return kwargs

    def form_valid(self, form: Form) -> HttpResponse:
        form.instance.project = self.project  # type: ignore
        return super().form_valid(form)

    def get_form_valid_message(self) -> str:
        link = cast(ProjectRelatedLink, self.object)
        return f'Created Related Link "{link.title}" for project "{link.project.machine_name}"'

    def get_success_url(self) -> str:
        link = cast(ProjectRelatedLink, self.object)
        logger.info(
            'project.relatedlink.create.success project=%s link_id=%s link_title=%s link_url=%s',
            link.project.machine_name,
            link.id,
            link.title,
            link.uri
        )
        return reverse('sphinx_hosting:project--update', args=[link.project.machine_name])


class ProjectRelatedLinkUpdateView(
    LoginRequiredMixin,
    PermissionRequiredMixin,
    FormValidMessageMixin,
    BaseUpdateView
):
    form_class = ProjectRelatedLinkUpdateForm
    model: Type[Model] = ProjectRelatedLink
    permission_required: str = 'sphinxhostingcore.change_project'

    def get_form_kwargs(self) -> Dict[str, str]:
        kwargs = super().get_form_kwargs()
        kwargs['project_machine_name'] = self.object.project.machine_name
        return kwargs

    def get_form_valid_message(self) -> str:
        return f'Updated Related Link "{self.object.title}" for project "{self.object.project.machine_name}"'

    def get_success_url(self) -> str:
        logger.info(
            'project.relatedlink.update.success project=%s link_id=%s link_title=%s link_url=%s',
            self.object.project.machine_name,
            self.object.id,
            self.object.title,
            self.object.uri
        )
        return reverse('sphinx_hosting:project--update', args=[self.object.project.machine_name])


class ProjectRelatedLinkDeleteView(
    LoginRequiredMixin,
    PermissionRequiredMixin,
    DeleteView
):
    model: Type[Model] = ProjectRelatedLink
    permission_required: str = 'sphinxhostingcore.change_project'

    def delete(self, request: HttpRequest, *args, **kwargs) -> HttpResponse:
        response = super().delete(request, *args, **kwargs)
        logger.info(
            'project.relatedlink.delete.success project=%s link_id=%s link_title=%s link_url=%s',
            self.object.project.machine_name,
            self.object.id,
            self.object.title,
            self.object.uri
        )
        return response

    def get_success_url(self) -> str:
        return reverse('sphinx_hosting:project--update', args=[self.object.project.machine_name])


# ===========================
# Versions
# ===========================


class VersionDetailView(
    LoginRequiredMixin,
    WildewidgetsMixin,
    SphinxHostingMenuMixin,
    DetailView
):
    """
    This view handles displaying the details page for a
    :py:class:`sphinx_hosting.models.Project` and handles updates to the project
    settings itself.
    """

    menu_item: str = "Projects"

    model: Type[Model] = Version
    slug_field: str = 'version'
    slug_url_kwarg: str = 'version'

    def get_object(self, queryset: QuerySet = None) -> Version:
        """
        If ``version`` is ``latest``, return the latest version of the
        :py:class:`sphinx_hosting.models.Project` whose ``machine_name`` is
        ``project_slug``.  Otherwise, return the version identified by
        ``version``.

        Raises:
            Http404: if the project has no versions, or the ``project_slug`` does
                not match any :py:class:`sphinx_hosting.models.Project` objects.

        Returns:
            The latest version of the project identified by ``project_slug``
        """
        if queryset is None:
            queryset = self.get_queryset()
        if self.kwargs['version'] == 'latest':
            project_slug = self.kwargs.get('project_slug', None)
            project = get_object_or_404(Project, machine_name=project_slug)
            version = project.latest_version
            if not version:
                raise Http404(f'Project "{project_slug}" has no versions')
            return version
        return super().get_object(queryset=queryset)

    def get_queryset(self) -> QuerySet[Version]:
        """
        Pre-filter our default queryset so that we only are able to get
        :py:class:`sphinx_hosting.models.Version` objects associated with the
        :py:class:`sphinx_hosting.models.Project` identified by our
        ``project_slug`` URLPath kwarg.

        Parameters:
            project_slug: the :py:attr:`sphinx_hosting.models.Project.machine_name`
                of our project

        Returns:
            A queryset of ``Version`` objects filtered by ``project_slug``
        """
        project_machine_name = self.kwargs.get('project_slug', None)
        return super().get_queryset().filter(project__machine_name=project_machine_name)

    def get_content(self) -> Widget:
        """
        Generate the set of widgets for this page.

        Returns:
            A populated page layout
        """
        layout = WidgetListLayout(f'{self.object.project.title} {self.object.version}')
        layout.add_widget(VersionInfoWidget(self.object))
        layout.add_widget(VersionSphinxPageTableWidget(version_id=self.object.pk))
        layout.add_widget(VersionSphinxImageTableWidget(version_id=self.object.pk))
        if self.object.head:
            layout.add_sidebar_link_button(
                'Read Docs',
                reverse(
                    'sphinx_hosting:sphinxpage--detail',
                    args=[
                        self.object.project.machine_name,
                        self.object.version,
                        self.object.head.relative_path
                    ]
                ),
                color='primary',
                css_class='mb-3'
            )
        user = cast(AbstractUser, self.request.user)
        if user.has_perm('sphinxhostingcore.delete_version'):
            layout.add_sidebar_form_button(
                'Delete Version',
                reverse('sphinx_hosting:version--delete', args=[self.object.project.machine_name, self.object.version]),
                color='outline-secondary',
                confirm_text=_("Are you sure you want to delete this version?"),
            )
        return layout

    def get_breadcrumbs(self) -> SphinxHostingBreadcrumbs:
        breadcrumbs = SphinxHostingBreadcrumbs()
        breadcrumbs.add_breadcrumb(self.object.project.title, url=self.object.project.get_absolute_url())
        breadcrumbs.add_breadcrumb(self.object.version)
        return breadcrumbs


class VersionUploadView(
    LoginRequiredMixin,
    PermissionRequiredMixin,
    FormInvalidMessageMixin,
    FormValidMessageMixin,
    BaseFormView
):
    permission_required: str = 'sphinxhostingcore.add_version'
    form_class = VersionUploadForm

    def get_form_valid_message(self) -> str:
        version = cast(Version, self.version)
        return f'Uploaded version "{version.version}" to project "{version.project.title}"'

    def form_valid(self, form: Form):
        with tempfile.TemporaryDirectory() as tmpdir:
            fs = FileSystemStorage(tmpdir)
            filename = fs.save('docs.tar.gz', content=self.request.FILES['file'])  # type: ignore
            path = os.path.join(fs.location, filename)
            self.version: Optional[Version] = None
            try:
                self.version = SphinxPackageImporter().run(filename=path, force=True)
            except Project.DoesNotExist:
                logger.error('version.upload.failed.no-such-project')
                self.messages.error(
                    'The project for the your tarball was not found.  Ensure that '
                    'the "project" field in your Sphinx "conf.py" matches the machine name '
                    'of an existing project here.'
                )
            except Exception as e:
                logger.error('version.upload.failed.unknown, error=%s', str(e))
                os.rename(path, '/tmp/uploaded_file')
                raise
        version = cast(Version, self.version)
        logger.info(
            'version.upload.success project_id=%s project_title=%s version=%s',
            version.project.id,
            version.project.title,
            version.version
        )
        return super().form_valid(form)

    def get_success_url(self) -> str:
        return reverse('sphinx_hosting:project--update', args=[self.kwargs['slug']])


class VersionDeleteView(
    LoginRequiredMixin,
    PermissionRequiredMixin,
    DeleteView
):
    permission_required: str = 'sphinxhostingcore.delete_version'
    model: Type[Model] = Version
    slug_field: str = 'version'
    slug_url_kwarg: str = 'version'

    def get_queryset(self) -> QuerySet[Version]:
        """
        Pre-filter our default queryset so that we only are able to get
        :py:class:`sphinx_hosting.models.Version` objects associated with the
        :py:class:`sphinx_hosting.models.Project` identified by our
        ``project_slug`` URLPath kwarg.

        Parameters:
            project_slug: the :py:attr:`sphinx_hosting.models.Project.machine_name`
                of our project

        Returns:
            A queryset of ``Version`` objects filtered by ``project_slug``
        """
        project_machine_name = self.kwargs.get('project_slug', None)
        return super().get_queryset().filter(project__machine_name=project_machine_name)

    def delete(self, request: HttpRequest, *args, **kwargs) -> HttpResponse:
        response = super().delete(request, *args, **kwargs)
        logger.info("version.delete.success project=%s id=%s", self.object.machine_name, self.object.pk)
        return response

    def get_success_url(self) -> str:
        return reverse('sphinx_hosting:project--update', args=[self.kwargs['project_slug']])


# ===========================
# Pages
# ===========================

class SphinxPageDetailView(
    LoginRequiredMixin,
    WildewidgetsMixin,
    SphinxHostingMenuMixin,
    DetailView
):
    """
    This view handles displaying the actual contents of the
    :py:class:`SphinxPage`, with navigation, table of contents and the actual
    page content.
    """

    model: Type[Model] = SphinxPage
    slug_field: str = 'relative_path'
    slug_url_kwarg: str = 'path'

    def get_navbar(self) -> Navbar:
        navbar = super().get_navbar()
        globaltoc = SphinxPageGlobalTableOfContentsMenu.parse_obj(self.object.version)
        globaltoc.title = self.object.version.project.title
        navbar.add_to_menu_section(globaltoc)
        return navbar

    def get_menu_item(self) -> str:
        return self.request.path

    def get_queryset(self) -> QuerySet[SphinxPage]:
        """
        Pre-filter our default queryset so that we only are able to get
        :py:class:`sphinx_hosting.models.SphinxPage` objects associated with the
        :py:class:`sphinx_hosting.models.Version` identified by our ``project_slug``
        and ``version`` URLPath kwargs.

        If ``version`` is ``latest``, return the latest version of the
        :py:class:`sphinx_hosting.models.Project`.

        Parameters:
            project_slug: the :py:attr:`sphinx_hosting.models.Project.machine_name`
                of our project
            version: the :py:attr:`sphinx_hosting.models.Version.version` string
                of our version

        Raises:
            Http404: if ``version`` is ``latest`` and the project has no
                versions, or the ``project_slug`` does not match any
                :py:class:`sphinx_hosting.models.Project` objects.

        Returns:
            A queryset of ``SphinxPage`` objects filtered to a particular version of
            a particular project.
        """
        project_slug = self.kwargs.get('project_slug', None)
        version = self.kwargs.get('version', None)
        if version == 'latest':
            project = get_object_or_404(Project, machine_name=project_slug)
            version_obj = project.latest_version
            if not version_obj:
                raise Http404(f'Project "{project_slug}" has no versions')
            version = version_obj.version
        return super().get_queryset().filter(
            version__version=version,
            version__project__machine_name=project_slug
        )

    def get_content(self) -> Widget:
        return SphinxPageLayout(self.object)

    def get_breadcrumbs(self) -> SphinxHostingBreadcrumbs:
        """
        Return our breadcrumbs for this page::

            Home -> Project -> Version -> SphinxPage.title

        Returns:
            This page's breadcrumbs
        """
        breadcrumbs = SphinxHostingBreadcrumbs()
        breadcrumbs.add_breadcrumb(
            self.object.version.project.title,
            url=self.object.version.project.get_absolute_url()
        )
        breadcrumbs.add_breadcrumb(self.object.version.version, url=self.object.version.get_absolute_url())
        breadcrumbs.add_breadcrumb(self.object.title)
        return breadcrumbs


# ===========================
# Search Views
# ===========================

class GlobalSphinxPageSearchView(
    LoginRequiredMixin,
    MessageMixin,
    SphinxHostingMenuMixin,
    WildewidgetsMixin,
    SearchView,
):
    """
    This is the view that renders our search results.
    """
    query: Optional[str] = None
    queryset: SearchQuerySet

    def form_invalid(self, form: ModelSearchForm) -> HttpResponse:
        self.queryset = self.get_queryset()
        self.object_list = self.queryset
        self.query: str = None
        context = self.get_context_data()
        return self.render_to_response(context)

    def form_valid(self, form: ModelSearchForm) -> HttpResponse:
        self.queryset = form.search().filter(is_latest='true')
        self.facets: Dict[str, List[str]] = {}
        if project_id := self.request.GET.get('project_id', None):
            self.queryset = self.queryset.filter(project_id=project_id)
            self.facets['project_id'] = [project_id]
        if classifier_name := self.request.GET.get('classifiers', None):
            self.queryset = self.queryset.filter(classifiers=classifier_name)
            self.facets['classifiers'] = [classifier_name]
        self.object_list = self.queryset
        self.query = form.cleaned_data[self.search_field]
        context = self.get_context_data()
        return self.render_to_response(context)

    def get_content(self) -> Widget:
        return PagedSearchLayout(self.object_list, self.query, facets=self.facets)

    def get_breadcrumbs(self) -> SphinxHostingBreadcrumbs:
        """
        Return our breadcrumbs for this page::

            Home -> Project -> Version -> SphinxPage.title

        Returns:
            This page's breadcrumbs
        """
        breadcrumbs = SphinxHostingBreadcrumbs()
        breadcrumbs.add_breadcrumb('Search')
        breadcrumbs.add_breadcrumb(f'Query: "{self.query}"')
        return breadcrumbs



# Generated by Django 4.2.3 on 2023-07-26 18:06

from django.db import migrations, models
import django.db.models.deletion
import django_extensions.db.fields


class Migration(migrations.Migration):

    dependencies = [
        ('sphinxhostingcore', '0011_machinenamefield'),
    ]

    operations = [
        migrations.AlterField(
            model_name='sphinxpage',
            name='body',
            field=models.TextField(blank=True, help_text='The body for the page, extracted from the page JSON, and modified to suit us.  Some pages have no body.  The body is actually stored as a Django template.', verbose_name='Body'),
        ),
        migrations.CreateModel(
            name='ProjectRelatedLink',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', django_extensions.db.fields.CreationDateTimeField(auto_now_add=True, verbose_name='created')),
                ('modified', django_extensions.db.fields.ModificationDateTimeField(auto_now=True, verbose_name='modified')),
                ('title', models.CharField(help_text='The title for this link', max_length=100, verbose_name='Link Title')),
                ('url', models.URLField(help_text='The URL for this link', max_length=256, verbose_name='Link URL')),
                ('project', models.ForeignKey(help_text='The project to which this link is related', on_delete=django.db.models.deletion.CASCADE, related_name='related_links', to='sphinxhostingcore.project')),
            ],
            options={
                'verbose_name': 'project related link',
                'verbose_name_plural': 'project related links',
                'ordering': ('title',),
                'unique_together': {('project', 'url')},
            },
        ),
    ]



import os
import tempfile
from typing import List, Type, Optional

from django.core.files.storage import FileSystemStorage
from django.db.models import Model
from django_filters import rest_framework as filters
from django_filters.filters import Filter, CharFilter, NumberFilter
from rest_framework import viewsets, permissions, mixins
from rest_framework.decorators import action
from rest_framework.parsers import MultiPartParser
from rest_framework.response import Response
from rest_framework.request import Request
from rest_framework.views import APIView

from sphinx_hosting.models import (
    Classifier,
    Project,
    ProjectRelatedLink,
    Version,
    SphinxPage,
    SphinxImage,
)
from sphinx_hosting.importers import SphinxPackageImporter

from .permissions import ChangeProjectPermission, AddVersionPermission
from .serializers import (
    ClassifierSerializer,
    ProjectSerializer,
    ProjectRelatedLinkSerializer,
    VersionSerializer,
    VersionUploadSerializer,
    SphinxPageSerializer,
    SphinxImageSerializer,
)


class ClassifierFilter(filters.FilterSet):

    name: Filter = CharFilter(
        field_name="name",
        lookup_expr="icontains",
        help_text="Filter by classifier name [case insensitive, partial match]",
    )

    class Meta:
        model: Type[Model] = Classifier
        fields: List[str] = ["name"]


class ClassifierViewSet(viewsets.ModelViewSet):

    permission_classes = [
        permissions.IsAuthenticated,
        permissions.DjangoModelPermissions,
    ]
    serializer_class = ClassifierSerializer
    queryset = Classifier.objects.all()
    filterset_class = ClassifierFilter


class ProjectFilter(filters.FilterSet):

    title: Filter = CharFilter(
        lookup_expr="icontains",
        help_text="Filter by project title, [case insensitive, partial match]",
    )
    machine_name: Filter = CharFilter(
        lookup_expr="icontains",
        help_text="Filter by project machine name, [case insensitive, partial match]",
    )
    description: Filter = CharFilter(
        lookup_expr="icontains",
        help_text="Filter by project description, [case insensitive, partial match]",
    )
    classifier: Filter = CharFilter(
        field_name="classifiers__name",
        lookup_expr="icontains",
        help_text="Filter by project classifier name [case insensitive, partial match]]",
    )

    class Meta:
        model: Type[Model] = Project
        fields: List[str] = ["title", "machine_name", "description", "classifier"]


class ProjectViewSet(viewsets.ModelViewSet):

    permission_classes = [
        permissions.IsAuthenticated,
        permissions.DjangoModelPermissions,
    ]
    serializer_class = ProjectSerializer
    queryset = Project.objects.all()
    filterset_class = ProjectFilter

    @action(detail=True)
    def latest_version(self, request: Request, pk: Optional[int] = None) -> Response:
        project = self.get_object()
        serializer = VersionSerializer(
            project.latest_version, context={"request": request}
        )
        return Response(serializer.data)


class VersionFilter(filters.FilterSet):

    project: Filter = NumberFilter()
    project_title: Filter = CharFilter(
        field_name="project__title",
        lookup_expr="icontains",
        help_text="Filter by project title [case insensitive, partial match]",
    )
    project_machine_name: Filter = CharFilter(
        field_name="project__machine_name",
        lookup_expr="icontains",
        help_text="Filter by project machine name [case insensitive, partial match]",
    )
    project_classifier: Filter = CharFilter(
        field_name="project__classifiers__name",
        lookup_expr="icontains",
        help_text="Filter by project classifier name [case insensitive, partial match]",
    )
    version_number: Filter = CharFilter(
        lookup_expr="iexact",
        help_text="Filter by version number [case insensitive, exact match]",
    )
    sphinx_version: Filter = CharFilter(
        lookup_expr="istartswith",
        help_text="Filter by Sphinx version [case insensitive, partial match to start of string]",
    )
    archived: Filter = filters.BooleanFilter(help_text="Filter by archived status")

    class Meta:
        model: Type[Model] = Version
        fields: List[str] = [
            "project",
            "project_title",
            "project_machine_name",
            "project_classifier",
            "version",
            "version_number",
            "sphinx_version",
            "archived",
        ]


class VersionViewSet(
    mixins.RetrieveModelMixin,
    mixins.DestroyModelMixin,
    mixins.ListModelMixin,
    viewsets.GenericViewSet,
):
    """
    Users can get, list and delete :py:class:`sphinx_hosting.models.Version` objects,
    but they can't create or update them the normal Django way.
    """

    permission_classes = [
        permissions.IsAuthenticated,
        permissions.DjangoModelPermissions | ChangeProjectPermission,
    ]
    serializer_class = VersionSerializer
    queryset = Version.objects.all()
    filterset_class = VersionFilter


class VersionUploadView(APIView):
    """
    This is the view to use to upload our sphinx tarballs.  It uploads to a
    temporary directory that disappears at the end of this view.

    To upload a file, you must submit as form-data, with a single file key named
    ``file``, with the ``Content-Disposition`` header like so::

        Content-Disposition: attachment;filename=yourdocs.tar.gz

    The filename you pass in the ``Content-Disposition`` header does not matter
    and is not used; set it to whatever you want.

    Example:

        To upload a file with ``curl`` to the endpoint for this view::

            curl \\
                -XPOST \\
                -H "Authorization: Token __THE_API_TOKEN__" \\
                -F 'file=@path/to/yourdocs.tar.gz' \\
                https://sphinx-hosting.example.com/api/v1/version/import/
    """

    serializer_class = VersionUploadSerializer
    permission_classes = [
        permissions.IsAuthenticated,
        AddVersionPermission
    ]
    parser_classes = (MultiPartParser,)

    def post(self, request: Request) -> Response:
        serializer = VersionUploadSerializer(data=request.data)
        serializer.is_valid(raise_exception=True)
        with tempfile.TemporaryDirectory() as tmpdir:
            fs = FileSystemStorage(tmpdir)
            filename = fs.save("docs.tar.gz", content=serializer.validated_data["file"])
            path = os.path.join(fs.location, filename)
            try:
                version = SphinxPackageImporter().run(filename=path, force=True)
            except Project.DoesNotExist as e:
                return Response(
                    {"status": "error", "message": str(e)},
                )
            except Exception:
                os.rename(path, "/tmp/uploaded_file")
                raise

        data = {
            "status": "success",
            "version_id": version.id,
            "project_id": version.project.id,
        }
        return Response(data, status=200)


class SphinxPageFilter(filters.FilterSet):

    project: Filter = NumberFilter(
        field_name="version__project", help_text="Filter by project ID"
    )
    project_title: Filter = CharFilter(
        field_name="version__project__title",
        lookup_expr="icontains",
        help_text="Filter by project title [case insensitive, partial match]",
    )
    project_machine_name: Filter = CharFilter(
        field_name="version__project__machine_name",
        lookup_expr="icontains",
        help_text="Filter by project machine name [case insensitive, partial match]",
    )
    project_classifier: Filter = CharFilter(
        field_name="version__project__classifiers__name",
        lookup_expr="icontains",
        help_text="Filter by project classifier name [case insensitive, partial match]",
    )
    version: Filter = NumberFilter(help_text="Filter by version ID")
    version_number: Filter = CharFilter(
        field_name="version__version",
        lookup_expr="iexact",
        help_text="Filter by version number [case insensitive, exact match]",
    )
    archived: Filter = filters.BooleanFilter(
        field_name="version__archived", help_text="Filter by archived status"
    )
    title: Filter = CharFilter(
        lookup_expr="icontains",
        help_text="Filter by page title [case insensitive, partial match]",
    )
    relative_path: Filter = CharFilter(
        lookup_expr="icontains",
        help_text="Filter by page relative path [case insensitive, partial match]",
    )
    sphinx_version: Filter = CharFilter(
        field_name="version__sphinx_version",
        lookup_expr="istartswith",
        help_text="Filter by Sphinx version [case insensitive, partial match to start of string]",
    )

    class Meta:
        model: Type[Model] = SphinxPage
        fields: List[str] = [
            "project",
            "project_title",
            "project_machine_name",
            "project_classifier",
            "version",
            "version_number",
            "archived",
            "title",
            "relative_path",
            "sphinx_version",
        ]


class SphinxPageViewSet(viewsets.ReadOnlyModelViewSet):
    """
    This is a read-only model set for
    :py:class:`sphinx_hosting.models.SphinxPage` models.  It is purposely
    read-only because we only want to update pages in the source Sphinx project,
    not here in the database.

    Even for our derived fields that we built out of the source, pages have a
    lot of interdependencies that need to be accounted for while editing.
    """

    permission_classes = [permissions.IsAuthenticated]
    serializer_class = SphinxPageSerializer
    queryset = SphinxPage.objects.all()
    filterset_class = SphinxPageFilter


class SphinxImageFilter(filters.FilterSet):

    project: Filter = NumberFilter(
        field_name="version__project", help_text="Filter by project ID"
    )
    project_title: Filter = CharFilter(
        field_name="version__project__title",
        lookup_expr="icontains",
        help_text="Filter by project title [case insensitive, partial match]",
    )
    project_machine_name: Filter = CharFilter(
        field_name="version__project__machine_name",
        lookup_expr="icontains",
        help_text="Filter by project machine name [case insensitive, partial match]",
    )
    project_classifier: Filter = CharFilter(
        field_name="version__project__classifiers__name",
        lookup_expr="icontains",
        help_text="Filter by project classifier name [case insensitive, partial match]",
    )
    version: Filter = NumberFilter(help_text="Filter by version ID")
    version_number: Filter = CharFilter(
        field_name="version__version",
        lookup_expr="iexact",
        help_text="Filter by version number [case insensitive, exact match]",
    )
    archived: Filter = filters.BooleanFilter(
        field_name="version__archived", help_text="Filter by archived status"
    )
    sphinx_version: Filter = CharFilter(
        field_name="version__sphinx_version",
        lookup_expr="istartswith",
        help_text="Filter by Sphinx version [case insensitive, partial match to start of string]",
    )
    orig_path: Filter = CharFilter(
        lookup_expr="icontains",
        help_text="Filter by original path [case insensitive, partial match]",
    )

    class Meta:
        model: Type[Model] = SphinxImage
        fields: List[str] = [
            "project",
            "version",
            "sphinx_version",
            "archived",
            "project_title",
            "project_machine_name",
            "project_classifier",
            "version_number",
            "orig_path",
        ]


class SphinxImageViewSet(viewsets.ReadOnlyModelViewSet):
    """
    This is a read-only model set for
    :py:class:`sphinx_hosting.models.SphinxImage` models.  It is purposely
    read-only because images are dependent objects of
    :py:class:`sphinx_hosting.models.SphinxPage` instances, and it makes no
    sense to update them independently.
    """

    permission_classes = [permissions.IsAuthenticated]
    serializer_class = SphinxImageSerializer
    queryset = SphinxImage.objects.all()
    filterset_class = SphinxImageFilter


class ProjectRelatedLinkFilter(filters.FilterSet):

    title: Filter = CharFilter(
        lookup_expr="icontains",
        help_text="Filter by link title, [case insensitive, partial match]",
    )
    project_title: Filter = CharFilter(
        field_name="project__title",
        lookup_expr="icontains",
        help_text="Filter by project title, [case insensitive, partial match]",
    )
    project_machine_name: Filter = CharFilter(
        field_name="project__machine_name",
        lookup_expr="icontains",
        help_text="Filter by project machine name, [case insensitive, partial match]",
    )
    project_description: Filter = CharFilter(
        field_name="project__description",
        lookup_expr="icontains",
        help_text="Filter by project description, [case insensitive, partial match]",
    )
    project_classifier: Filter = CharFilter(
        field_name="project__classifiers__name",
        lookup_expr="icontains",
        help_text="Filter by project classifier name [case insensitive, partial match]]",
    )

    class Meta:
        model: Type[Model] = ProjectRelatedLink
        fields: List[str] = [
            "title",
            "project_title",
            "project_machine_name",
            "project_description",
            "project_classifier"
        ]


class ProjectRelatedLinkViewSet(viewsets.ModelViewSet):

    permission_classes = [
        permissions.IsAuthenticated,
        ChangeProjectPermission
    ]
    serializer_class = ProjectRelatedLinkSerializer
    queryset = ProjectRelatedLink.objects.all()
    filterset_class = ProjectRelatedLinkFilter



from argparse import ArgumentParser
from pprint import pprint

from django.core.management.base import BaseCommand
from django.utils.text import slugify

from sphinx_hosting.models import Version
from sphinx_hosting.models import SphinxGlobalTOCHTMLProcessor


class Command(BaseCommand):
    """
    **Usage**: ``./manage.py print_globaltoc <project_machine_name> <version number>``

    Print the global table of contents for a :py:class:`Version`.
    """
    args = '<project_machine_name> <version_number>'
    help = 'Print the deduced page hierarchy for a version of a project.'

    def add_arguments(self, parser: ArgumentParser) -> None:
        parser.add_argument(
            'project',
            metavar='project_machine_name',
            type=str,
            help='The machine_name of the project.'
        )
        parser.add_argument(
            'version',
            metavar='version_number',
            type=str,
            help='The version number for which we want to print a page tree.'
        )
        parser.add_argument(
            'verbose',
            action='store_true',
            default=False,
            help='Increase verbosity.'
        )

    def handle(self, *args, **options) -> None:
        try:
            version = Version.objects.get(
                project__machine_name=slugify(options['project']),
                version=options['version']
            )
        except Version.DoesNotExist as e:
            self.stdout.write(self.style.ERROR(str(e)))
            return

        pprint(SphinxGlobalTOCHTMLProcessor().run(version, verbose=options['verbose']))



from django.apps import AppConfig
from django.utils.translation import gettext_lazy as _


class SphinxHostingAppConfig(AppConfig):
    name: str = "sphinx_hosting"
    label: str = "sphinxhostingcore"
    verbose_name = _("Sphinx Hosting core")
    default_auto_field: str = "django.db.models.AutoField"



# Generated by Django 4.2.3 on 2023-07-27 17:59

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('sphinxhostingcore', '0012_project_relatedlinks'),
    ]

    operations = [
        migrations.RenameField(
            model_name='projectrelatedlink',
            old_name='url',
            new_name='uri',
        ),
        migrations.AlterUniqueTogether(
            name='projectrelatedlink',
            unique_together={('project', 'uri')},
        ),
    ]



from typing import Any

from django.core.exceptions import ValidationError
from django.core.validators import RegexValidator
from django.utils.deconstruct import deconstructible
from django.utils.html import strip_tags
from django.utils.regex_helper import _lazy_re_compile
from django.utils.translation import gettext_lazy as _


@deconstructible
class NoHTMLValidator:
    """
    Raises a ValidationError if the given value contains any HTML.
    """
    message: str = 'Cannot contain any HTML'
    code: str = 'invalid'

    def __call__(self, value: str) -> None:
        if not value == strip_tags(value):
            raise ValidationError(self.message, code=self.code)

    def __eq__(self, other: Any) -> bool:
        return (
            isinstance(other, NoHTMLValidator)
        )


machine_name_re = _lazy_re_compile(r"^[-a-zA-Z0-9_.]+\Z")
validate_machine_name = RegexValidator(
    machine_name_re,
    # Translators: "letters" means latin letters: a-z and A-Z.
    _("Enter a valid “machine name” consisting of letters, numbers, underscores, hyphens or periods."),
    "invalid",
)



# Generated by Django 4.1.2 on 2022-11-22 19:27

from django.db import migrations, models
import django.db.models.deletion
import django_extensions.db.fields
import sphinx_hosting.models
import sphinx_hosting.validators


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='Project',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', django_extensions.db.fields.CreationDateTimeField(auto_now_add=True, verbose_name='created')),
                ('modified', django_extensions.db.fields.ModificationDateTimeField(auto_now=True, verbose_name='modified')),
                ('title', models.CharField(help_text='The human name for this project', max_length=100, verbose_name='Project Name')),
                ('description', models.CharField(blank=True, help_text='A brief description of this project', max_length=256, null=True, validators=[sphinx_hosting.validators.NoHTMLValidator()], verbose_name='Brief Description')),
                ('machine_name', models.SlugField(help_text='Must be unique.  Set this to the value of "project" in Sphinx\'s. conf.py', unique=True, verbose_name='Machine Name')),
            ],
            options={
                'verbose_name': 'project',
                'verbose_name_plural': 'projects',
            },
        ),
        migrations.CreateModel(
            name='SphinxPage',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', django_extensions.db.fields.CreationDateTimeField(auto_now_add=True, verbose_name='created')),
                ('modified', django_extensions.db.fields.ModificationDateTimeField(auto_now=True, verbose_name='modified')),
                ('relative_path', models.CharField(help_text='The path to the page under our top slug', max_length=255, verbose_name='Relative page path')),
                ('content', models.TextField(help_text='The full JSON payload for the page', verbose_name='Content')),
                ('title', models.CharField(help_text='Just the title for the page, extracted from the page JSON', max_length=255, verbose_name='Title')),
                ('body', models.TextField(blank=True, help_text='Just the body for the page, extracted from the page JSON. Some pages have no body.', verbose_name='Body')),
                ('local_toc', models.TextField(blank=True, default=None, help_text='Table of Contents for headings in this page', null=True, verbose_name='Local Table of Contents')),
                ('next_page', models.OneToOneField(help_text='The next page in the documentation set', null=True, on_delete=django.db.models.deletion.CASCADE, related_name='previous_page', to='sphinxhostingcore.sphinxpage')),
                ('parent', models.ForeignKey(help_text='The parent page of this page', null=True, on_delete=django.db.models.deletion.CASCADE, related_name='children', to='sphinxhostingcore.sphinxpage')),
            ],
            options={
                'verbose_name': 'sphinx page',
                'verbose_name_plural': 'sphinx pages',
            },
        ),
        migrations.CreateModel(
            name='Version',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', django_extensions.db.fields.CreationDateTimeField(auto_now_add=True, verbose_name='created')),
                ('modified', django_extensions.db.fields.ModificationDateTimeField(auto_now=True, verbose_name='modified')),
                ('version', models.CharField(help_text='A version of our project', max_length=64, verbose_name='Version')),
                ('sphinx_version', models.CharField(blank=True, default=None, help_text='The version of Sphinx used to documentation for this set', max_length=64, null=True, verbose_name='Sphinx Version')),
                ('head', models.OneToOneField(help_text='The top page of the documentation set for this version of our project', null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='+', to='sphinxhostingcore.sphinxpage')),
                ('project', models.ForeignKey(help_text='The Project to which this Version belong', on_delete=django.db.models.deletion.CASCADE, related_name='versions', to='sphinxhostingcore.project')),
            ],
            options={
                'get_latest_by': 'modified',
                'abstract': False,
            },
        ),
        migrations.AddField(
            model_name='sphinxpage',
            name='version',
            field=models.ForeignKey(help_text='The Version to which this page belongs', on_delete=django.db.models.deletion.CASCADE, related_name='pages', to='sphinxhostingcore.version'),
        ),
        migrations.AlterUniqueTogether(
            name='sphinxpage',
            unique_together={('version', 'relative_path')},
        ),
        migrations.CreateModel(
            name='SphinxImage',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', django_extensions.db.fields.CreationDateTimeField(auto_now_add=True, verbose_name='created')),
                ('modified', django_extensions.db.fields.ModificationDateTimeField(auto_now=True, verbose_name='modified')),
                ('orig_path', models.CharField(help_text='The original path to this file in the Sphinx documentation package', max_length=256, verbose_name='Original Path')),
                ('file', models.FileField(help_text='The actual image file', upload_to=sphinx_hosting.models.sphinx_image_upload_to, verbose_name='An image file')),
                ('version', models.ForeignKey(help_text='The version of our project documentation with which this image is associated', on_delete=django.db.models.deletion.CASCADE, related_name='images', to='sphinxhostingcore.version')),
            ],
            options={
                'verbose_name': 'sphinx image',
                'verbose_name_plural': 'sphinx images',
                'unique_together': {('version', 'orig_path')},
            },
        ),
    ]



# Generated by Django 4.1.2 on 2023-01-19 20:01

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion
import django_extensions.db.fields
import sphinx_hosting.validators


class Migration(migrations.Migration):

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ('sphinxhostingcore', '0005_orig_global_toc'),
    ]

    operations = [
        migrations.CreateModel(
            name='Classifier',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(help_text='The classifier spec for this classifier, e.g. "Language :: Python"', max_length=255, unique=True, verbose_name='Classifier Name')),
            ],
            options={
                'verbose_name': 'classifier',
                'verbose_name_plural': 'classifiers',
            },
        ),
        migrations.RemoveField(
            model_name='version',
            name='global_toc',
        ),
        migrations.AlterField(
            model_name='project',
            name='machine_name',
            field=models.SlugField(help_text='Must be unique.  Set this to the slugified value of "project" in Sphinx\'s. conf.py', unique=True, verbose_name='Machine Name'),
        ),
        migrations.AlterField(
            model_name='version',
            name='project',
            field=models.ForeignKey(help_text='The Project to which this Version belongs', on_delete=django.db.models.deletion.CASCADE, related_name='versions', to='sphinxhostingcore.project'),
        ),
        migrations.AlterField(
            model_name='version',
            name='sphinx_version',
            field=models.CharField(blank=True, default=None, help_text='The version of Sphinx used to create this documentation set', max_length=64, null=True, verbose_name='Sphinx Version'),
        ),
        migrations.AlterField(
            model_name='version',
            name='version',
            field=models.CharField(help_text='The version number for this release of the Project', max_length=64, verbose_name='Version'),
        ),
        migrations.CreateModel(
            name='ProjectPermissionGroup',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', django_extensions.db.fields.CreationDateTimeField(auto_now_add=True, verbose_name='created')),
                ('modified', django_extensions.db.fields.ModificationDateTimeField(auto_now=True, verbose_name='modified')),
                ('name', models.CharField(help_text='The name for this permission group', max_length=100, unique=True, verbose_name='Permission Group Name')),
                ('description', models.CharField(blank=True, help_text='A brief description of this permission group', max_length=256, null=True, validators=[sphinx_hosting.validators.NoHTMLValidator()], verbose_name='Brief Description')),
                ('users', models.ManyToManyField(related_name='project_permission_groups', to=settings.AUTH_USER_MODEL)),
            ],
            options={
                'verbose_name': 'project permission group',
                'verbose_name_plural': 'project permission groups',
            },
        ),
        migrations.AddField(
            model_name='project',
            name='classifiers',
            field=models.ManyToManyField(related_name='projects', to='sphinxhostingcore.classifier'),
        ),
        migrations.AddField(
            model_name='project',
            name='permission_groups',
            field=models.ManyToManyField(related_name='projects', to='sphinxhostingcore.projectpermissiongroup'),
        ),
    ]



from django.core.management.base import BaseCommand
from rich.tree import Tree
from rich import print as rich_print

from sphinx_hosting.models import Classifier, ClassifierNode


class TreePrinter:
    """
    Parse the tree of :py:class:`sphinx_hosting.models.ClassifierNode` objects
    we get from :py:meth:`sphinx_hosting.models.ClassifierManager.tree` and
    print them to stdout in a pretty format.
    """

    def __init__(self) -> None:
        self.classifier_tree = Classifier.objects.tree()
        self.tree: Tree = Tree('Classifiers')
        for node in self.classifier_tree.values():
            branch = self.tree.add(self.title(node))
            self.build(branch, node)

    def title(self, node: ClassifierNode) -> str:
        if node.classifier:
            return f'{node.title} (ID: {node.classifier.id})'
        return node.title

    def build(self, branch: Tree, node: ClassifierNode) -> None:
        if node.items:
            for child in node.items.values():
                leaf = branch.add(self.title(child))
                self.build(leaf, child)

    def print(self) -> None:
        rich_print(self.tree)


class Command(BaseCommand):
    """
    **Usage**: ``./manage.py print_classifier_tree``

    Print the :py:class:`Classifier` hierarchy.
    """
    help = ('Print the Classifier hierarchy.')

    def handle(self, *args, **options) -> None:
        TreePrinter().print()



from dataclasses import dataclass, field
from pathlib import Path
import re
from typing import cast, Any, List, Dict, Optional
from urllib.parse import urlparse, unquote
from django.conf import settings
from django.db import models
from django.utils.translation import gettext_lazy as _
from django.utils.functional import cached_property
from django.urls import reverse
from django_extensions.db.models import TimeStampedModel
from lxml import etree
import lxml.html
from lxml.html import HtmlElement
from wildewidgets.models import ViewSetMixin

from .fields import MachineNameField
from .settings import MAX_GLOBAL_TOC_TREE_DEPTH
from .validators import NoHTMLValidator


F = models.Field
M2M = models.ManyToManyField
FK = models.ForeignKey


# --------------------------
# Dataclasses
# --------------------------

@dataclass
class TreeNode:
    """
    This is a :py:class:`dataclass` that we use with :py:class:`SphinxPageTree`
    to build out the global navigation structure for a set of documentation for
    a :py:class:`Version`.
    """

    #: The page title
    title: str
    #: This page
    page: Optional["SphinxPage"] = None
    #: The :py:class:`SphinxPage` after this one
    prev: Optional["SphinxPage"] = None
    #: The :py:class:`SphinxPage` before this one
    next: Optional["SphinxPage"] = None
    #: The :py:class:`SphinxPage` that is this page's parent
    parent: Optional["SphinxPage"] = None
    #: The :py:class:`TreeNode` objects that are this page's children
    children: List["TreeNode"] = field(default_factory=list)

    @classmethod
    def from_page(cls, page: "SphinxPage") -> "TreeNode":
        """
        Build a :py:class:`TreeNode` from ``page``.

        Note:
            This does not populate :py:attr:`children`; :py:class:`SphinxPageTree`
            will populate it as appropriate as it ingests pages.

        Args:
            page: the :py:class:`SphinxPage` from which to build a node

        Returns:
            A configured node.
        """
        return cls(
            page=page,
            title=page.title,
            next=page.next_page,
            prev=page.previous_page.first(),
            parent=page.parent
        )


@dataclass
class ClassifierNode:

    title: str
    classifier: Optional["Classifier"] = None
    items: Dict[str, "ClassifierNode"] = field(default_factory=dict)


# --------------------------
# Helper classes
# --------------------------

class SphinxPageTree:
    """
    A class that holds the page hierarchy for the set of :py:class:`SphinxPage`
    pages in a :py:class:`Version`.as a linked set of :py:class:`TreeNode`
    objects.

    The page heirarchy is built by starting at :py:attr:`Version.head` and
    following the page linkages by looking at :py:attr:`SphinxPage.next_page`,
    stopping the traversal when we find a :py:attr:`SphinxPage.next_page` that is ``None``.

    As we traverse, if a :py:attr:`SphinxPage.parent` is not ``None``, find the
    :py:class:`TreeNode` for that parent, and add the page to
    :py:attr:`TreeNode.children`.

    For pages who have no :py:attr:`SpinxPage.parent`, assume they are top level
    children of the set, and make them children of :py:class:`Version.head`.

    Load it like so::

        >>> project = Project.objects.get(machine_name='my-project')
        >>> version = project.versions.get(version='1.0.0')
        >>> tree = SphinxPageTree(version)

    You can then traverse the built hierarchy by starting at
    :py:attr:`SphinxPageTree.head`, looking at its children, then looking at
    their children, etc..

        >>>

    """

    def __init__(self, version: "Version"):
        #: The :py:class:`Version that this tree examines
        self.version: Version = version
        self.nodes: Dict[int, TreeNode] = {}
        self.nodes[self.version.head.id] = TreeNode.from_page(self.version.head)
        #: The top page in the page hierarchy
        self.head: TreeNode = self.nodes[self.version.head.id]
        self.build(version)

    def build(self, version: "Version"):
        self.add_page(version.head)

    def add_page(self, page: "SphinxPage"):
        node = TreeNode.from_page(page)
        self.nodes[page.id] = node
        if node.parent:
            if node.parent.id in self.nodes:
                self.nodes[node.parent.id].children.append(node)
        else:
            # The top level pages that are not head will not have any parent
            # because Sphinx doesn't think that way
            if node != self.head:
                self.head.children.append(node)
        if node.next:
            self.add_page(node.next)

    def _traverse_level(self, pages, nodes):
        for node in nodes:
            if node.page:
                pages.append(node.page)
                if node.children:
                    self._traverse_level(pages, node.children)

    def traverse(self) -> List["SphinxPage"]:
        """
        Return a list of the pages represented in this tree.
        """
        pages: List["SphinxPage"] = [cast("SphinxPage", self.head.page)]
        self._traverse_level(pages, self.head.children)
        return pages


class SphinxPageTreeProcessor:

    def build_item(self, node: TreeNode) -> Dict[str, Any]:
        """
        Build a :py:class:`wildewdigets.MenuItem` compatible
        dict representing ``node``.

        Args:
            node: the current node in our page tree

        Returns:
            A dict suitable for loading into a :py:class:`wildewidgets.MenuItem`.
        """
        item: Dict[str, Any] = {
            'text': node.title,
            'url': None,
            'icon': None,
            'items': []
        }
        if node.page:
            item['url'] = node.page.get_absolute_url()
        return item

    def build(self, items: List[Dict[str, Any]], node: TreeNode) -> None:
        """
        Build a :py:class:`wildewdigets.MenuItem` compatible
        dict representing ``node``, and append it to ``items``.

        if ``node`` has children, recurse into those children, building out
        our submenus.

        Args:
            items: the current list of ``MenuItem`` compatible dicts for the
                current level of the menu
            node: the current node in our page tree
        """
        item = self.build_item(node)
        if node.children:
            for child in node.children:
                self.build(item['items'], child)
        items.append(item)

    def run(self, version: "Version") -> List[Dict[str, Any]]:
        """
        Parse the :py:func:`Version.page_tree` and return a struct that works
        with
        :py:class:`sphinx_hosting.wildewidgets.SphinxPageGlobalTableOfContentsMenu.parse_obj`

        The returned struct should look something like this::

            [
                {'text': 'foo'},
                {'text': 'bar', 'url': '/foo', 'icon': None}
                {'text': 'bar', 'url': '/foo', 'icon': None, items: [{'text': 'blah' ...} ...]}
                ...
            ]

        Args:
            version: the version whose global table of contents we are parsing

        Returns:
            A list of dicts representing the global menu structure
        """
        self.sphinx_tree = version.page_tree
        items: List[Dict[str, Any]] = []
        items.append(self.build_item(self.sphinx_tree.head))
        for child in self.sphinx_tree.head.children:
            self.build(items, child)
        return items


class SphinxGlobalTOCHTMLProcessor:
    """
    **Usage**: ``SphinxGlobalTOCHTMLProcessor().run(version, globaltoc_html)```

    This importer is used to parse the ``globaltoc`` key in JSON output of
    Sphinx pages built with the `sphinxcontrib-jsonglobaltoc
    <https://github.com/caltechads/sphinxcontrib-jsonglobaltoc>`_ extension.

    Sphinx uses your ``.. toctree:`` declarations in your ``.rst`` files to
    build site navigation for your document tree, and
    ``sphinxcontrib-jsonglobaltoc`` saves the Sphinx HTML produced by those
    ``..toctree`` as the ``globaltoc`` key in the `.fjson` output.

    Note:

        Sphinx ``.. toctree:`` are ad-hoc -- they're up to how the author wants
        to organize their content, and may not reflect how files are filled out
        in the filesystem.
    """

    def __init__(self, max_level: int = 2) -> None:
        #: Generate at most this many levels of menus and submenus
        self.max_level: int = max_level

    def fix_href(self, href: str) -> str:
        p = urlparse(unquote(href))
        url = reverse(
            'sphinx_hosting:sphinxpage--detail',
            kwargs={
                'project_slug': self.version.project.machine_name,
                'version': self.version.version,
                'path': p.path.strip('/')
            }
        )
        if p.fragment:
            url += f'#{p.fragment}'
        return url

    def parse_ul(self, html: HtmlElement, level: int = 1) -> List[Dict[str, Any]]:
        """
        Process ``html``, an ``lxml`` parsed set of elements representing the contents
        of a ``<ul>`` from a Sphinx table of contents and return a list of
        :py:class:`sphinx_hosting.wildewidgets.MenuItem` objects.

        Any ``href`` in links found will be converted to its full
        ``django-sphinx-hosting`` path.

        If we find another ``<ul>`` inside ``html``, process it by passing its
        contents to :py:meth:`parse_ul` again, incrementing the menu level.

        If ``level`` is greater than :py:attr:`max_level`, return an empty list,
        stopping our recursion.

        Args:
            html: the list of elements that are the contents of the parent ``<ul>``

        Keyword Args:
            level: the current menu level

        Returns:
            The ``<ul>`` contents as a list of dicts
        """
        items: List[Dict[str, Any]] = []
        if level <= self.max_level:
            for li in html.iterchildren():
                item: Dict[str, Any] = {
                    'text': 'placeholder',
                    'url': None,
                    'icon': None,
                    'items': []
                }
                for elem in li:
                    if elem.tag == 'a':
                        item['text'] = elem.text_content()
                        item['url'] = self.fix_href(elem.attrib['href'])
                    if elem.tag == 'ul':
                        item['items'].extend(self.parse_ul(elem, level=level + 1))
                items.append(item)
        return items

    def parse_globaltoc(self, html: HtmlElement) -> List[Dict[str, Any]]:
        """
        Parse our global table of contents HTML blob and return a list of
        :py:class:`sphinx_hosting.wildewidgets.MenuItem` objects.

        Add a first node that points to the root doc, also.  The root doc can't
        add itself to its ``toctree`` blocks, so we need to do it ourselves.

        How our mapping works:

        * Multiple top level ``<ul>`` tags separated by ``<p class="caption">``
          tags will be merged into a single list.
        * ``<p class="caption ...">CONTENTS</p>`` becomes ``{'text': 'CONTENTS'}```
        * Any ``href`` will be converted to its full ``django-sphinx-hosting`` path

        Args:
            version: the version whose global table of contents we are parsing
            html: the lxml parsed HTML of the global table of contents from
                Sphinx
        """
        root_url = reverse(
            'sphinx_hosting:sphinxpage--detail',
            kwargs={
                'project_slug': self.version.project.machine_name,
                'version': self.version.version,
                'path': self.version.head.relative_path
            }
        )
        items: List[Dict[str, Any]] = [
            {
                'text': 'Home',
                'url': root_url,
                'icon': None,
                'items': []
            }
        ]
        for elem in html.iterchildren():
            if elem.tag == 'p' and 'caption' in elem.classes:
                # Captions only appear at the top level, even if you assign
                # captions in your toctree declaration in your sub levels with
                # :caption:, so we only process them here.
                items.append({'text': elem.text_content()})
            if elem.tag == 'ul':
                items.extend(self.parse_ul(elem))
        return items

    def run(
        self,
        version: "Version",
        verbose: bool = False
    ) -> List[Dict[str, Any]]:
        """
        Parse the global table of contents found as
        ``version.head.orig_global_toc`` into a data struct suitable for use
        with
        :py:class:`sphinx_hosting.wildewidgets.SphinxPageGlobalTableOfContentsMenu.parse_obj`
        and return it.

        How our mapping works:

        * Multiple top level ``<ul>`` tags separated by ``<p class="caption">``
          tags will be merged into a single list.
        * ``<p class="caption ...">CONTENTS</p>`` becomes ``{'text': 'CONTENTS'}```
        * Any ``href`` for links found will be converted to its full
          ``django-sphinx-hosting`` path

        The returned struct should look something like this::

            [
                {'text': 'foo'},
                {'text': 'bar', 'url': '/project/version/foo', 'icon': None}
                {'text': 'bar', 'url': '/project/version/bar', 'icon': None, items: [{'text': 'blah' ...} ...]}
                ...
            ]

        Args:
            version: the version whose global table of contents we are parsing

        Keyword Args:
            verbose: if ``True``, pretty print the HTML of the globaltoc

        Returns:
            A list of dicts representing the global menu structure
        """
        self.version = version
        if self.version.head.orig_global_toc:
            # This is really only necessary to get the pretty printer below to
            # work properly.  If the \n chars are not removed, lxml won't indent
            # properly
            global_toc_html = re.sub(r'\n', '', self.version.head.orig_global_toc)
            html = lxml.html.fromstring(global_toc_html)
            if verbose:
                print(
                    etree.tostring(  # pylint: disable=c-extension-no-member
                        html,
                        method='xml',
                        encoding='unicode',
                        pretty_print=True
                    )
                )
            return self.parse_globaltoc(html)
        else:
            return []

# --------------------------
# FileField upload functions
# --------------------------


def sphinx_image_upload_to(instance: "SphinxImage", filename: str) -> str:
    """
    Set the upload path within our ``MEDIA_ROOT`` for any images used by our Sphinx
    documentation to be::

        {project machine_name}/{version}/images/{image basename}

    Args:
        instance: the :py:class:`SphinxImage` object
        filename: the original path to the file

    Returns:
        The properly formatted path to the file
    """
    path = Path(instance.version.project.machine_name) / Path(instance.version.version) / 'images'
    path = path / Path(filename).name
    return str(path)

# --------------------------
# Managers
# --------------------------


class ClassifierManager(models.Manager):

    def tree(self) -> Dict[str, ClassifierNode]:
        """
        Given our classifiers, which are ``::`` separated lists of terms
        like::

            Section :: Subsection :: Name
            Section :: Subsection :: Name2
            Section :: Subsection :: Name3
            Section :: Subsection

        Return a tree-like data structure that looks like::

            {
                'Section': ClassifierNode(
                    title='Section'
                    items={
                        'Subsection': ClassifierNode(
                            title='Subsection',
                            classifier=Classifier(name="Section :: Subsection"),
                            items: {
                                'Name': ClassifierNode(
                                    title='Name',
                                    classifier=Classifier(
                                        name='Section :: Subsection :: Name'
                                    )
                                ),
                                ...
                            }
                        )
                    }
                )
            }
        """
        nodes: Dict[str, ClassifierNode] = {}
        current: Optional[ClassifierNode] = None
        for classifier in self.get_queryset().all():
            parts = classifier.name.split(' :: ')
            if parts[0] not in nodes:
                nodes[parts[0]] = ClassifierNode(title=parts[0])
            current = nodes[parts[0]]
            if len(parts) > 1:
                for part in parts[1:]:
                    if part not in current.items:
                        current.items[part] = ClassifierNode(title=part)
                    current = current.items[part]
            current.classifier = classifier
        return nodes


# --------------------------
# Models
# --------------------------


class Classifier(ViewSetMixin, models.Model):
    """
    A :py:class:`Project` can be tagged with one or more :py:class:`Classifier`
    tags.  This allows you to group projects by ecosystem, or type, for example.

    Use `PyPI <www.pypi.org>`_ classifiers as an example of how to use a single
    field for classifying across many dimensions.

    Examples::

        Ecosystem :: CMS
        Language :: Python
        Owner :: DevOps :: AWS
    """
    objects = ClassifierManager()

    name: F = models.CharField(
        'Classifier Name',
        help_text=_('The classifier spec for this classifier, e.g. "Language :: Python"'),
        max_length=255,
        unique=True,
    )

    def save(self, *args, **kwargs) -> None:
        """
        Overrides :py:meth:`django.db.models.Model.save`.

        Override save to create any missing classifiers in our chain.  For example,
        if we want to create this classifier::

            Foo :: Bar :: Baz

        But ``Foo :: Bar`` does not yet exist in the database, create that before
        creating ``Foo :: Bar :: Baz``.  We do this so that when we filter our projects
        by classifier, we can filter by ``Foo :: Bar`` and ``Foo :: Bar :: Baz``.
        """
        parts = [p.strip() for p in self.name.split('::')]
        if len(parts) > 2:
            name = parts[0]
            for part in parts[1:-1]:
                name = f'{name} :: {part}'
                if not self.objects.filter(name=name).exists():
                    new_classifiier = Classifier(name=name)
                    new_classifiier.save(using=kwargs.get('using', settings.DEFAULT_DB_ALIAS))
        # Rejoin our parts to ensure we always get a classifier that looks like
        # "part :: part :: part" instead of "part:: part::part"
        self.name = ' :: '.join(parts)
        super().save(*args, **kwargs)

    def __str__(self) -> str:
        return str(self.name)

    class Meta:
        verbose_name = _('classifier')
        verbose_name_plural = _('classifiers')


class ProjectPermissionGroup(
    ViewSetMixin,
    TimeStampedModel,
    models.Model
):
    """
    A :py:class:`Project` can be assigned to one or more
    :py:class:`ProjectPermissionGroup` groups.  This restricts viewing of the
    project to users which belong to those groups.

    Examples::

        Ecosystem :: CMS
        Language :: Python
        Owner :: DevOps :: AWS
    """
    name: F = models.CharField(
        'Permission Group Name',
        help_text=_('The name for this permission group'),
        max_length=100,
        unique=True,
    )
    description: F = models.CharField(
        'Brief Description',
        max_length=256,
        null=True,
        blank=True,
        help_text=_('A brief description of this permission group'),
        validators=[NoHTMLValidator()]
    )

    users: M2M = models.ManyToManyField(
        settings.AUTH_USER_MODEL,
        related_name='project_permission_groups'
    )

    class Meta:
        verbose_name = _('project permission group')
        verbose_name_plural = _('project permission groups')


class Project(ViewSetMixin, TimeStampedModel, models.Model):
    """
    A Project is what a set of Sphinx docs describes: an application, a library, etc.

    Projects have versions (:py:class:`Version`) and versions have Sphinx pages
    (:py:class:`SphinxPage`).
    """

    title: F = models.CharField(
        'Project Name',
        help_text=_('The human name for this project'),
        max_length=100
    )
    description: F = models.CharField(
        'Brief Description',
        max_length=256,
        null=True,
        blank=True,
        help_text=_('A brief description of this project'),
        validators=[NoHTMLValidator()]
    )
    machine_name: F = MachineNameField(
        'Machine Name',
        unique=True,
        help_text=_(
            """Must be unique.  Set this to the slugified value of "project" in Sphinx's. conf.py"""
        )
    )

    permission_groups: M2M = models.ManyToManyField(
        ProjectPermissionGroup,
        related_name='projects',
    )
    classifiers: M2M = models.ManyToManyField(
        Classifier,
        related_name='projects',
    )

    def __str__(self) -> str:  # pylint: disable=invalid-str-returned
        return self.title

    @property
    def latest_version(self) -> "Optional[Version]":
        """
        Return the latest version (by version number) of our project
        documentation, if any.

        Returns:
            The latest version of our project.
        """
        return self.versions.order_by('-version').first()

    def get_absolute_url(self) -> str:
        return reverse('sphinx_hosting:project--detail', args=[self.machine_name])

    def get_update_url(self) -> str:
        return reverse('sphinx_hosting:project--update', args=[self.machine_name])

    def get_latest_version_url(self) -> Optional[str]:
        if self.latest_version:
            return reverse(
                'sphinx_hosting:sphinxpage--detail',
                args=[
                    self.machine_name,
                    self.latest_version.version,
                    self.latest_version.head.relative_path
                ]
            )
        return None

    class Meta:
        verbose_name = _('project')
        verbose_name_plural = _('projects')


class Version(TimeStampedModel, models.Model):
    """
    A ``Version`` is a specific version of a :py:class:`Project`.  Versions own
    :py:class:`SphinxPage` objects.
    """

    project: FK = models.ForeignKey(
        Project,
        on_delete=models.CASCADE,
        related_name='versions',
        help_text=_('The Project to which this Version belongs'),
    )
    version: F = models.CharField(
        'Version',
        max_length=64,
        null=False,
        help_text=_('The version number for this release of the Project'),
    )
    sphinx_version: F = models.CharField(
        'Sphinx Version',
        max_length=64,
        null=True,
        blank=True,
        default=None,
        help_text=_('The version of Sphinx used to create this documentation set')
    )
    archived: F = models.BooleanField(
        'Archived?',
        default=False,
        help_text=_(
            'Whether this version should be excluded from search indexes'
        )
    )

    head: FK = models.OneToOneField(
        "SphinxPage",
        on_delete=models.SET_NULL,
        null=True,
        related_name='+',  # disable our related_name for this one
        help_text=_('The top page of the documentation set for this version of our project'),
    )

    def __str__(self) -> str:
        return f'{self.project.title}-{self.version}'

    @property
    def is_latest(self) -> bool:
        return self == self.project.latest_version

    @property
    def page_tree(self) -> SphinxPageTree:
        """
        Return the page hierarchy for the set of :py:class:`SphinxPage` pages in
        this version.

        The page hierarchy is build by traversing the pages in the set, starting
        with :py:attr:`head`.

        Returns:
            The page hierarchy for this version.
        """
        return SphinxPageTree(self)

    def mark_searchable_pages(self) -> None:
        """
        Set the :py:attr:`SphinxPage.searchable` flag on
        the searchable pages in this version.

        Searchable pages are ones that:

        * Are not in :py:attr:`SphinxPage.SPECIAL_PAGES`
        * Do not have a part of their relative path that starts with ``_``.

        Go through the pages in this version, and set
        :py:attr:`SphinxPage.searchable` to ``True`` for all those which meet the above
        requirements, ``False`` otherwise.
        """
        ignored_paths = list(SphinxPage.SPECIAL_PAGES.keys())
        for page in self.pages.all():
            if (
                page.relative_path not in ignored_paths and
                not page.relative_path.startswith('_') and
                '/_' not in page.relative_path
            ):
                page.searchable = True
            else:
                page.searchable = False
            page.save()

    @cached_property
    def globaltoc(self) -> Dict[str, List[Dict[str, Any]]]:
        """
        Build a struct that looks like this::

            {
                items: [
                    {'text': 'foo'},
                    {'text': 'bar', 'url': '/foo', 'icon': None}
                    {'text': 'bar', 'url': '/foo', 'icon': None, items: [{'text': 'blah' ...} ...]}
                    ...
                ]
            }

        suitable for constructing a
        :py:class:`sphinx_hosting.wildewidgets.SphinxPageGlobalTableOfContentsMenu`
        for this :py:class:`Version`.

        """
        items = SphinxGlobalTOCHTMLProcessor(max_level=MAX_GLOBAL_TOC_TREE_DEPTH).run(self)
        if not items:
            items = SphinxPageTreeProcessor().run(self)
        return {'items': items}

    def purge_cached_globaltoc(self) -> None:
        """
        Purge the cached output from our :py:meth:`globaltoc` property.
        """
        try:
            del self.globaltoc
        except AttributeError:
            # This means self.globaltoc hasn't been accessed yet
            pass

    def get_absolute_url(self) -> str:
        return reverse(
            'sphinx_hosting:version--detail',
            args=[self.project.machine_name, self.version]
        )

    def save(self, *args, **kwargs):
        """
        Overriding :py:meth:`django.db.models.Model.save` here so that we can
        purge our cached global table of contents.
        """
        super().save(*args, **kwargs)
        self.purge_cached_globaltoc()


class SphinxPage(TimeStampedModel, models.Model):
    """
    A ``SphinxPage`` is a single page of a set of Sphinx documentation.
    ``SphinxPage`` objects are owned :py:class:`Version` objects, which are in
    turn owned by :py:class:`Project` objects.
    """

    #: This is a mapping between filename and title that identifies the
    #: special pages that Sphinx produces on its own and gives them
    #: reasonable titles.  These pages have no ``title`` key in their
    #: json data, but ``title`` is required for pages
    SPECIAL_PAGES: Dict[str, str] = {
        'genindex': 'General Index',
        'py-modindex': 'Module Index',
        'np-modindex': 'Module Index',
        'search': 'Search',
        '_modules/index': 'Module code'
    }

    version: FK = models.ForeignKey(
        Version,
        on_delete=models.CASCADE,
        related_name='pages',
        help_text=_('The Version to which this page belongs'),
    )
    relative_path: F = models.CharField(
        'Relative page path',
        help_text=_('The path to the page under our top slug'),
        max_length=255
    )
    content: F = models.TextField(
        'Content',
        help_text=_('The full JSON payload for the page')
    )
    title: F = models.CharField(
        'Title',
        max_length=255,
        help_text=_('Just the title for the page, extracted from the page JSON')
    )
    orig_body: F = models.TextField(
        'Body (Original)',
        blank=True,
        help_text=_(
            'The original body for the page, extracted from the page JSON. Some pages have no body. '
            'We save this here in case we need to reprocess the body at some later date.'
        ),
    )
    body: F = models.TextField(
        'Body',
        blank=True,
        help_text=_(
            'The body for the page, extracted from the page JSON, and modified to suit us.  '
            'Some pages have no body.  The body is actually stored as a Django template.'
        ),
    )
    orig_local_toc: F = models.TextField(
        'Local Table of Contents (original)',
        blank=True,
        null=True,
        default=None,
        help_text=_(
            'The original table of contents for headings in this page.'
            'We save this here in case we need to reprocess the table of contents '
            'at some later date.'
        ),
    )
    local_toc: F = models.TextField(
        'Local Table of Contents',
        blank=True,
        null=True,
        default=None,
        help_text=_('Table of Contents for headings in this page, modified to work in our templates'),
    )
    orig_global_toc: F = models.TextField(
        'Global Table of Contents (original)',
        blank=True,
        null=True,
        default=None,
        help_text=_(
            'The original global table of contents HTML attached to this page, if any.  This '
            'will only be present if you had "sphinxcontrib-jsonglobaltoc" installed in your "extensions" '
            'in the Sphinx conf.py'
        ),
    )
    searchable: F = models.BooleanField(
        'Searchable',
        default=False,
        help_text=_('Should this page be included in the search index?')
    )

    parent: FK = models.ForeignKey(
        "SphinxPage",
        on_delete=models.CASCADE,
        null=True,
        related_name="children",
        help_text=_('The parent page of this page'),
    )

    # This has to be a ForeignKey here and not a OneToOneField becuase
    # more than one page can have the same next_page.
    next_page: FK = models.ForeignKey(
        "SphinxPage",
        on_delete=models.CASCADE,
        null=True,
        related_name="previous_page",
        help_text=_('The next page in the documentation set'),
    )

    def __str__(self) -> str:  # pylint: disable=invalid-str-returned
        return f'{self.version.project.title}-{self.version.version}: {self.relative_path}'

    def get_absolute_url(self) -> str:
        return reverse(
            'sphinx_hosting:sphinxpage--detail',
            args=[
                self.version.project.machine_name,
                self.version.version,
                self.relative_path
            ]
        )

    class Meta:
        verbose_name = _('sphinx page')
        verbose_name_plural = _('sphinx pages')
        unique_together = ('version', 'relative_path')


class SphinxImage(TimeStampedModel, models.Model):
    """
    A ``SphinxImage`` is an image file referenced in a Sphinx document.  When
    importing documenation packages, we extract all images from the package,
    upload them into Django storage and update the Sphinx HTML in
    :py:attr:`SphinxPage.body` to reference the URL for the uploaded image
    instead of its original url.
    """

    version: FK = models.ForeignKey(
        Version,
        on_delete=models.CASCADE,
        related_name='images',
        help_text=_('The version of our project documentation with which this image is associated')
    )
    orig_path: F = models.CharField(
        _('Original Path'),
        max_length=256,
        help_text=_('The original path to this file in the Sphinx documentation package')
    )
    file: F = models.FileField(
        _('An image file'),
        upload_to=sphinx_image_upload_to,
        help_text=_('The actual image file')
    )

    class Meta:
        verbose_name = _('sphinx image')
        verbose_name_plural = _('sphinx images')
        unique_together = ('version', 'orig_path')


class ProjectRelatedLink(TimeStampedModel, models.Model):
    """
    A ``ProjectRelatedLink`` is a link to an external resource that is related
    to a :py:class:`Project`.
    """

    title: F = models.CharField(
        'Link Title',
        max_length=100,
        help_text=_('The title for this link')
    )
    uri: F = models.URLField(
        'Link URL',
        max_length=256,
        help_text=_('The URL for this link')
    )
    project: FK = models.ForeignKey(
        Project,
        on_delete=models.CASCADE,
        related_name='related_links',
        help_text=_('The project to which this link is related')
    )

    def get_update_url(self) -> str:
        return reverse("sphinx_hosting:projectrelatedlink--update", kwargs={'pk': self.pk})

    def get_delete_url(self) -> str:
        return reverse("sphinx_hosting:projectrelatedlink--delete", kwargs={'pk': self.pk})

    class Meta:
        verbose_name = _('project related link')
        verbose_name_plural = _('project related links')
        unique_together = ('project', 'uri')
        ordering = ('title',)



# Generated by Django 4.1.2 on 2022-12-08 23:32

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('sphinxhostingcore', '0002_original_html'),
    ]

    operations = [
        migrations.AddField(
            model_name='version',
            name='global_toc',
            field=models.TextField(blank=True, default=True, help_text='The global table of contents for this version, if any.  This is HTML, and will only be present if you had "sphinxcontrib-jsonglobaltoc" installed in your "extensions" in the Sphinx conf.py', null=True, verbose_name='Global Table of Contents'),
        ),
    ]



# Generated by Django 4.1.2 on 2022-11-22 23:23

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('sphinxhostingcore', '0001_initial'),
    ]

    operations = [
        migrations.AddField(
            model_name='sphinxpage',
            name='orig_body',
            field=models.TextField(blank=True, help_text='The original body for the page, extracted from the page JSON. Some pages have no body.We save this here in case we need to reprocess the body at some later date.', verbose_name='Body (Original'),
        ),
        migrations.AddField(
            model_name='sphinxpage',
            name='orig_local_toc',
            field=models.TextField(blank=True, default=None, help_text='The original table of contents for headings in this page.We save this here in case we need to reprocess the table of contents at some later date.', null=True, verbose_name='Local Table of Contents (original)'),
        ),
        migrations.AlterField(
            model_name='sphinxpage',
            name='body',
            field=models.TextField(blank=True, help_text='The body for the page, extracted from the page JSON, and modified to suit us.  Some pages have no body.', verbose_name='Body'),
        ),
        migrations.AlterField(
            model_name='sphinxpage',
            name='local_toc',
            field=models.TextField(blank=True, default=None, help_text='Table of Contents for headings in this page, modified to work in our templates', null=True, verbose_name='Local Table of Contents'),
        ),
    ]



# Generated by Django 4.1.2 on 2023-01-06 00:23

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('sphinxhostingcore', '0004_next_page_FK'),
    ]

    operations = [
        migrations.AddField(
            model_name='sphinxpage',
            name='orig_global_toc',
            field=models.TextField(blank=True, default=None, help_text='The original global table of contents HTML attached to this page, if any.  This will only be present if you had "sphinxcontrib-jsonglobaltoc" installed in your "extensions" in the Sphinx conf.py', null=True, verbose_name='Global Table of Contents (original)'),
        ),
    ]



from dataclasses import dataclass
import io
import json
from pathlib import Path
import re
import tarfile
from typing import Any, Dict, IO, List, Optional, cast
import urllib.parse


from django.utils.text import slugify
import lxml.html
from lxml.etree import XML  #: pylint: disable=no-name-in-module

from .exc import VersionAlreadyExists
from .logging import logger
from .models import Project, Version, SphinxPage, SphinxImage

ImageMap = Dict[str, SphinxImage]


@dataclass
class PageTreeNode:
    """
    A data structure to temporarily hold relationships between
    :py:class:`sphinx_hosting.models.SphinxPage` objects while importing pages.

    In the page JSON we get from Sphinx, we only know the titles of related
    pages, so we store them here along with the
    :py:class:`sphinx_hosting.models.SphinxPage` we created from our JSON, and
    then do another pass through these :py:class:`PageTreeNode` objects to link
    our pages together.

    This is used in :py:meth:`SphinxPackageImporter.link_pages`.
    """

    #: The page for this node
    page: SphinxPage
    #: The title of the parent page for this node, if any
    parent_title: Optional[str] = None
    #: The title of the next page for this node, if any
    next_title: Optional[str] = None


class SphinxPackageImporter:
    """
    **Usage**: ``SphinxPackageImporter().run(sphinx_tarfilename)```

    Import a tarfile of a built set of Sphinx documentation into the database.

    .. important::

        Before importing, there must be a
        :py:class:`sphinx_hosting.models.Project` in the database whose
        ``machine_name`` matches the ``project`` in Sphinx's ``conf.py`` config
        file for the docs to be imported.

    The documentation package should have been built via the ``json`` output from
    ``sphinx-build``, so either::

        make json

    or::

        sphinx-build -n -b json build/json

    The tarfile should be built like so::

        cd build
        tar zcf mydocs.tar.gz json

    ensuring that the package contents are enclosed in a folder.

    When run, :py:class:`SphinxPackageImporter` will look inside the tarfile at
    the ``globalcontext.json`` file to determine which project and version we should
    associate these pages with.

    * The ``project`` key in will be used to look up the
      :py:class:`sphinx_hosting.models.Project` to associate these Sphinx pages
      with, using ``project`` as :py:attr:`sphinx_hosting.models.Project.machine_name`
    * The ``version`` key will be used to create a new
      :py:class:`sphinx_hosting.models.Version` object tied to that project

    Once the :py:class:`sphinx_hosting.models.Version` has been created, the
    pages in the tarfile will be created as
    :py:class:`sphinx_hosting.models.SphinxPage` objects, and the images will be
    created as :py:class:`sphinx_hosting.models.SphinxImage` objects.
    """

    # Sometimes pages have weird titles -- replace them with their filename
    ODD_TITLES: List[str] = [
        '&lt;no title&gt;'
    ]

    def __init__(self) -> None:
        #: Used to map original Sphinx image paths to our Django storage path
        self.image_map: ImageMap = {}
        #: Used to link pages to their parent pages, and to their next pages
        self.page_tree: Dict[str, PageTreeNode] = {}
        self.name_map: Dict[str, str] = {}
        #: the contents of globalcontext.json
        self.config: Dict[str, Any] = {}

    def _get_file(self, package: tarfile.TarFile, filename: str) -> io.BufferedReader:
        """
        Look through the member names in our tarfile ``package`` for
        ``filename``, and return an open file descriptor on that file.

        Note:
          We have to do it this way instead of using ``package.getmember(name)``
          because we don't know the name of the containing folder.

        Args:
            package: the opened Sphinx documentation tarfile
            filename: the name of the file we're looking for

        Raises:
            KeyError: no file named ``filename`` was found in the tarfile

        Returns:
            The opened file descriptor for our file.
        """
        if not self.name_map:
            self.name_map = {str(Path(*Path(name).parts[1:])): name for name in package.getnames()}
        if filename not in self.name_map:
            raise KeyError(f'Sphinx docs TarFile has no file named "{filename}"')
        return cast(io.BufferedReader, package.extractfile(self.name_map[filename]))

    def _update_image_src(self, body: str) -> str:
        """
        Given an HTML body of a Sphinx page, update the ``<img src="path">``
        references to template tag expressions that load the actual image URL
        from the :py:class:`sphinx_hosting.models.SphinxImage` objects at render time.

        We need to defer filling in the URL of the image until render time
        because of things like storing images in S3 and using time-limited S3
        auth parameters to retrieve the image from a private bucket.  Those
        parameters expire typically after an hour, so if we don't defer figuring
        out the URL for our images, we end up storing a stale URL.

        Also deal with any lightboxes by converting them to the appropriate form
        to work with Tabler lightboxes.

        Args:
            body: the HTML body of a Sphinx document

        Returns:
            ``body`` with its ``<img>`` urls and lightbox attributes updated
        """
        if not body:
            return ''
        html = lxml.html.fromstring(body)
        images = html.cssselect('img')
        for image in images:
            src = re.sub(r'\.\./', '', image.attrib['src'])
            if src in self.image_map:
                image.attrib['src'] = f'{{% sphinximage_url {self.image_map[src].id} %}}'

        # also deal with any lightbox <a>
        lightboxes = html.cssselect('a[data-lightbox]')
        for lightbox in lightboxes:
            lightbox.attrib['data-fslightbox'] = lightbox.attrib['data-lightbox']
            del lightbox.attrib['data-lightbox']
            if 'data-title' in lightbox.attrib:
                lightbox.attrib['data-caption'] = lightbox.attrib['data-title']
                del lightbox.attrib['data-title']
            src = re.sub(r'\.\./', '', lightbox.attrib['href'])
            if src in self.image_map:
                lightbox.attrib['href'] = f'{{% sphinximage_url {self.image_map[src].id} %}}'
        return lxml.html.tostring(html).decode('utf-8')

    def load_config(self, package: tarfile.TarFile) -> None:
        """
        Load the ``globalcontext.json`` file for later reference.

        Args:
            package: the opened Sphinx documentation tarfile
        """
        self.config = json.loads(self._get_file(package, 'globalcontext.json').read())

    def get_version(self, package: tarfile.TarFile, force: bool = False) -> Version:
        """
        Look in ``package`` for a member named ``globalcontext.json``, and load
        that file as JSON.

        Extract these things from that JSON:

            * the version string from the ``release`` key.
            * the ``machine_name`` of the :py:class:`Project` for this
              documentation tarfile as the slugified version of the ``project``
              key

        Return a new :py:class:`Version` instance on the project.

        Args:
            package: the opened Sphinx documentation tarfile

        Keyword Args:
            force: if ``True``, re-use an existing version, purging any docs and
              images associated with it first

        Raises:
            Project.DoesNotExist: no :py:class:`Project` exists whose
                ``machine_name`` matches the slugified ``project`` setting
                in the Sphinx package's ``conf.py``
            VersionAlreadyExists: a :py:class:`Version` with version string
                ``release`` from the Sphinx ``conf.py`` already exists for our
                project, and ``force`` was not ``True``
        """
        machine_name = self.config['project']
        project = Project.objects.get(machine_name=machine_name)
        v = project.versions.filter(version=self.config['release']).first()
        if v:
            if not force:
                raise VersionAlreadyExists(
                    f"""Version {self.config['release']} of Project(machine_name="{machine_name}") """
                    """already exists."""
                )
            v.pages.all().delete()
            v.images.all().delete()
            v.sphinx_version = self.config['sphinx_version']
            v.head = None
        else:
            v = Version(
                project=project,
                version=self.config['release'],
                sphinx_version=self.config['sphinx_version']
            )
            v.save()
        return v

    def import_images(self, package: tarfile.TarFile, version: Version) -> None:
        """
        Import all images in our Sphinx documentation into the database before
        importing any pages, then return a lookup dict for doing ``<img
        src="image_path">`` replacements in the page bodies.

        Args:
            package: the opened Sphinx documentation tarfile
            version: the :py:class:`Version` which which to associate our images
        """
        for member in package.getmembers():
            path = Path(*Path(member.name).parts[1:])
            if path.match('_images/*'):
                fd = package.extractfile(member)
                orig_path: str = str(path)
                image = SphinxImage(version=version, orig_path=orig_path)
                image.file.save(orig_path, fd)
                image.save()
                self.image_map[orig_path] = image
                logger.info(
                    "%s.image.imported project=%s version=%s orig_path=%s url=%s id=%s",
                    self.__class__.__name__,
                    version.project.machine_name,
                    version.version,
                    image.orig_path,
                    image.file.url,
                    image.id
                )

    def _fix_page_title(self, path: str, data: Dict[str, Any]) -> None:
        """
        Ensure that there is a ``title`` key in ``data``, the JSON data from our
        .fjson file.  Some special pages don't have a ``title`` key in their
        JSON data, so we supply one based on their filename, or by copying
        another key from ``data``.

        Args:
            path: the file path in the tarfile data: the JSON data from our file
        """
        if 'title' not in data:
            data['title'] = 'UNKNOWN'
        if path in SphinxPage.SPECIAL_PAGES:
            data['title'] = SphinxPage.SPECIAL_PAGES[path]
        if data['title'] in self.ODD_TITLES:
            data['title'] = path
        if 'title' not in data:
            # Some of the special pages don't have 'title' keys
            if 'indextitle' in data:
                data['title'] = data['indextitle']
            else:
                data['title'] = SphinxPage.SPECIAL_PAGES[path]

    def _fix_link_hrefs(self, path: str, body: str) -> str:
        """
        Given an HTML body of a Sphinx page, update the ``<a href="path">``
        references for "path" to be rendered at page render time.  If we don't
        do this, a lot of links won't work, because they do
        index page, instead of being relative to the root of the docs, and won't
        work.

        Args:
            path: the path to the current page
            body: the HTML body of a Sphinx document

        Returns:
            ``body`` with its ``<a>`` urls and updated
        """
        if not body:
            return ''

        # Parse the HTML body into an lxml tree
        html = lxml.html.fromstring(body)

        # Find all internal references
        links = html.cssselect('a.reference.internal')

        # For each link, update its URL to be rendered at page render time
        for link in links:
            href = link.attrib['href']
            anchor = None
            if '#' in href:
                href, anchor = href.split('#')
            if href.endswith('/'):
                href = href[:-1]
            # To deal with relative links, we need to know our current path
            # and then compute the absolute path from that.
            levels = href.count('../')
            if levels:
                href = re.sub('^(../)*', '', href)
                href = '/'.join(path.split('/')[:-(levels)] + [href])
            link.attrib['href'] = "{{% url 'sphinx_hosting:sphinxpage--detail' project_slug='{}' version='{}' path='{}' %}}".format(  # noqa:E501  # pylint: disable=line-too-long
                self.config['project'],
                self.config['release'],
                href
            )
            if anchor:
                link.attrib['href'] += f'#{anchor}'

        # Return the updated HTML body
        return lxml.html.tostring(html).decode('utf-8')

    def _fix_page_body(self, path: str, data: Dict[str, Any]) -> None:
        """
        Do any work needed to prepare the page body before inserting into the
        database.  This means:

        * Ensure the ``body`` key exists in ``data``
        * Update the ``img`` sources to point to our Django storage location.
          We uploaded them to our storage during :py:meth:`import_images`.
        * Update the ``href``s for any ``<a>`` links to be rendered at page
          render time.
        * Update the ``<table>``s to have the CSS classes we need for them to
          display nicely.

        Args:
            path: the path to the current page
            data: the JSON data from our file
        """
        if 'body' not in data or data['body'] is None:
            # Ensure we always have data['body'] defined as a string, for when
            # we create the SphinxPage, below
            data['body'] = ''
        data['orig_body'] = data['body']
        if data['body']:
            # Update the img src for any images in data['body'] for to point to our
            # Django storage locations
            data['body'] = self._update_image_src(data['body'])
            # Update the hrefs for any <a> links to be absolute.  The relative
            # paths we get from Sphinx end up being relative to the Sphinx index
            # document instead of to the root of the docs
            data['body'] = self._fix_link_hrefs(path, data['body'])
            html = lxml.html.fromstring(data['body'])
            # remove the first <h1> -- we'll display the page title another way
            first_h1 = html.cssselect('h1')
            if first_h1:
                first_h1[0].getparent().remove(first_h1[0])
            # Fix our tables to look better
            tables = html.cssselect('table')
            for table in tables:
                wrapper = XML('<div class="table-responsive"></div>')
                parent = table.getparent()
                parent.append(wrapper)
                wrapper.insert(0, table)
                table.classes.add('table')
                table.classes.add('table-striped')
                table.classes.add('border')
                for tr in table.cssselect('thead > tr'):
                    tr.classes.discard('row-even')
                    tr.classes.discard('row-odd')
                for tr in table.cssselect('th'):
                    tr.classes.discard('head')
                    tr.classes.add('p-2')
                for tr in table.cssselect('tbody > tr'):
                    tr.classes.discard('row-even')
                    tr.classes.discard('row-odd')
                for div in table.cssselect('tbody > tr div.line'):
                    div.classes.discard('line')
                    div.classes.add('text-start')
                for div in table.cssselect('tbody > tr p'):
                    div.classes.add('text-start')
            data['body'] = lxml.html.tostring(html).decode('utf-8')
            # Unescape our template tags after lxml has converted our {% %}
            # to entities.
            tags = [m.group() for m in re.finditer(r'%7B%%20.*?%20%%7D', data['body'])]
            for tag in tags:
                data['body'] = data['body'].replace(tag, urllib.parse.unquote(tag))
            # Convert the weird paragraph symbols to actual paragraph symbols
            data['body'] = re.sub(r'#61633;', r'para;', data['body'])

    def _fix_toc(self, data: Dict[str, Any]) -> None:
        """
        Update our page's local table of contents (``data['toc']`) to have the CSS
        classes we need in order for it to display properly.

        Args:
            data: the decoded JSON data of the sphinx page
        """
        if 'toc' not in data:
            return
        data['orig_toc'] = data['toc']
        html = lxml.html.fromstring(data['toc'])
        ul_first = html.cssselect('ul:first-child')[0]
        # Turn the first <ul> into a tabler vertical nav
        ul_first.classes.add('nav-vertical')
        # Turn all <uls> into nav-pills and nav
        for ul in html.cssselect('ul'):
            ul.classes.add('nav')
            ul.classes.add('nav-pills')
        # Make all list items into nav-items
        for li in html.cssselect('li'):
            li.classes.add('nav-item')
        # Make <a> into nav-links
        for link in html.cssselect('a'):
            link.classes.add('nav-link')
        # Now make the embedded uls collapsable
        for ul in html.cssselect('li > ul'):
            wrapper = XML('<div class="d-flex flex-row justify-content-between align-items-center"></div>')
            link = ul.getprevious()
            link.addprevious(wrapper)
            wrapper.insert(0, link)
            target = f'menu-{slugify(link.text_content())}'
            wrapper.append(XML(
                '<a class="toc__toggle nav-link-toggle" data-bs-toggle="collapse" '
                f'aria-expanded="false" data-bs-target="#{target}"></a>'
            ))
            ul.attrib['id'] = target
            ul.classes.add('collapse')
        try:
            link = html.cssselect('a:nth-child(2)')[0]
            link.attrib['aria-expanded'] = 'true'
        except IndexError:
            pass
        try:
            ul = html.cssselect('li:first-child ul')[0]
            ul.classes.add('show')
        except IndexError:
            pass
        data['toc'] = lxml.html.tostring(html).decode('utf-8')

    def _update_page_tree(
        self,
        page: SphinxPage,
        data: Dict[str, Any]
    ) -> None:
        """
        Update ``tree``, our page linkage tree, which we will use in
        :py:meth:`link_pages` to set :py:attr:`SphinxPage.parent` and
        :py:attr:`SphinxPage.next_page` appropriately.

        Args:
            tree: our page linkage tree
            data: the JSON data from our file
        """
        parent: Optional[str] = None
        if 'parents' in data and data['parents']:
            parent = [p['title'] for p in data['parents']][-1]
        next_title: Optional[str] = None
        if 'next' in data and data['next']:
            next_title = data['next']['title']
        self.page_tree[page.title] = PageTreeNode(
            page=page,
            parent_title=parent,
            next_title=next_title
        )

    def import_pages(self, package: tarfile.TarFile, version: Version) -> None:
        """
        Import a all pages from ``package`` into the database as
        :py:class:`sphinx_hosting.models.SphinxPage` objects, associating them
        with :py:class:`Version` ``version``.

        Args:
            package: the tarfile of the sphinx docs
            version: the :py:class:`Version` object to associated data

        Returns:
            The page linkage tree for consumption by :py:meth:`link_pages`.
        """
        for member in package.getmembers():
            path: str = str(Path(*Path(member.name).parts[1:]))
            if path.split('/')[-1].startswith('._'):
                # This is a Mac OS X AppleDouble hidden file.  Ignore it and
                # move on.  It just has MacOS specific metadata we don't care
                # about.
                continue
            if path.endswith('.fjson'):
                # files that contain page data will have a .fjson extension
                path = path.replace('.fjson', '')
                fd = cast(io.BufferedReader, package.extractfile(member))
                data = json.loads(fd.read())
                self._fix_page_title(path, data)
                self._fix_page_body(path, data)
                self._fix_toc(data)
                page = SphinxPage(
                    version=version,
                    relative_path=path,
                    content=json.dumps(data),
                    title=data['title'],
                    orig_body=data['orig_body'],
                    body=data['body'],
                    orig_local_toc=data['orig_toc'] if 'orig_toc' in data else None,
                    local_toc=data['toc'] if 'toc' in data else None,
                    orig_global_toc=data['globaltoc'] if 'globaltoc' in data else None
                )
                page.save()
                self._update_page_tree(page, data)
                logger.info(
                    "%s.page.imported project=%s version=%s relpath=%s title=%s id=%s",
                    self.__class__.__name__,
                    version.project.machine_name,
                    version.version,
                    page.relative_path,
                    page.title,
                    page.id
                )

    def link_pages(self) -> None:
        """
        Given :py:attr:`page_tree``, a list of page linkages (parent, next,
        prev), link all the :py:class:`sphinx_hosting.models.SphinxPage` objects
        in that list to their next page and their parent page.

        Args:
            tree: the page linkage tree
        """
        for link in self.page_tree.values():
            page = link.page
            logger.info(
                "%s.page.linking relpath=%s title=%s id=%s",
                self.__class__.__name__,
                page.relative_path,
                page.title,
                page.id
            )
            if link.parent_title:
                page.parent = self.page_tree[link.parent_title].page
                logger.info(
                    "%s.page.linked-parent relpath=%s title=%s parent=%s",
                    self.__class__.__name__,
                    page.relative_path,
                    page.title,
                    page.parent.title
                )
            if link.next_title:
                page.next_page = self.page_tree[link.next_title].page
                logger.info(
                    "%s.page.linked-next relpath=%s title=%s next=%s",
                    self.__class__.__name__,
                    page.relative_path,
                    page.title,
                    page.next_page.title
                )
            page.save()

    def run(
        self,
        filename: str = None,
        file_obj: IO = None,
        force: bool = False
    ) -> Version:
        """
        Load the pages in the tarfile identified by ``filename`` into
        the database as :py:class:`Version` ``version`` of :py:class:`Project`
        ``project``.  See the class docs for :py:class:`SphinxPackageImporter` for
        more background on how to prepare the package named by ``filename``.

        Args:
            filename: the filename of the gzipped tar archive of the Sphinx pages

        Keyword Args:
            force: if ``True``, overwrite the docs for an existing version

        Raises:
            Project.DoesNotExist: no :py:class:`Project` exists whose
                ``machine_name`` matches the slugified ``project`` setting
                in the Sphinx package's ``conf.py``
            VersionAlreadyExists: a :py:class:`Version` with version string
                ``release`` from the Sphinx package's ``conf.py``
                already exists for our project, and ``force`` was not ``True``
        """
        assert not all([filename, file_obj]), 'provide either "filename" or "file_obj" but not both'
        with tarfile.open(name=filename, fileobj=file_obj) as package:
            self.load_config(package)
            version = self.get_version(package, force=force)
            self.import_images(package, version)
            self.import_pages(package, version)
            self.link_pages()
        # Point version.head at the top page of the documentation set
        version.head = SphinxPage.objects.get(version=version, relative_path=self.config['root_doc'])
        version.save()
        # Mark the appropriate pages as indexable
        version.mark_searchable_pages()
        return version
